"""
The Results object is the result of running a survey. 
It is not typically instantiated directly, but is returned by the run method of a `Job` object.
"""

from __future__ import annotations
import json
import random
from collections import UserList, defaultdict
from typing import Optional, Callable, Any, Type, Union, List, TYPE_CHECKING

from bisect import bisect_left

from edsl.Base import Base
from edsl.exceptions.results import (
    ResultsError,
    ResultsBadMutationstringError,
    ResultsColumnNotFoundError,
    ResultsInvalidNameError,
    ResultsMutateError,
    ResultsFilterError,
    ResultsDeserializationError,
)

if TYPE_CHECKING:
    from edsl.surveys.Survey import Survey
    from edsl.data.Cache import Cache
    from edsl.agents.AgentList import AgentList
    from edsl.language_models.model import Model
    from edsl.scenarios.ScenarioList import ScenarioList
    from edsl.results.Result import Result
    from edsl.jobs.tasks.TaskHistory import TaskHistory
    from edsl.language_models.ModelList import ModelList
    from simpleeval import EvalWithCompoundTypes

from edsl.results.ResultsExportMixin import ResultsExportMixin
from edsl.results.ResultsGGMixin import ResultsGGMixin
from edsl.results.results_fetch_mixin import ResultsFetchMixin
from edsl.utilities.remove_edsl_version import remove_edsl_version

def ensure_fetched(method):
    """A decorator that checks if remote data is loaded, and if not, attempts to fetch it."""
    def wrapper(self, *args, **kwargs):
        if not self._fetched:
            # If not fetched, try fetching now.
            # (If you know you have job info stored in self.job_info)
            self.fetch_remote(self.job_info)
        return method(self, *args, **kwargs)
    return wrapper

def ensure_ready(method):
    """
    Decorator for Results methods.
    
    If the Results object is not ready, for most methods we return a NotReadyObject.
    However, for __repr__ (and other methods that need to return a string), we return
    the string representation of NotReadyObject.
    """
    from functools import wraps

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if self.completed:
            return method(self, *args, **kwargs)
        # Attempt to fetch remote data
        try:
            if hasattr(self, "job_info"):
                self.fetch_remote(self.job_info)
        except Exception as e:
            print(f"Error during fetch_remote in {method.__name__}: {e}")
        if not self.completed:
            not_ready = NotReadyObject(name = method.__name__, job_info = self.job_info)
            # For __repr__, ensure we return a string
            if method.__name__ == "__repr__" or method.__name__ == "__str__":
                return not_ready.__repr__()
            return not_ready
        return method(self, *args, **kwargs)

    return wrapper

class NotReadyObject:
    """A placeholder object that prints a message when any attribute is accessed."""
    def __init__(self, name: str, job_info: RemoteJobInfo):
        self.name = name
        self.job_info = job_info
        #print(f"Not ready to call {name}")

    def __repr__(self):
        message = f"""Results not ready - job still running on server."""
        for key, value in self.job_info.creation_data.items():
            message += f"\n{key}: {value}"
        return message

    def __getattr__(self, _):
        return self
    
    def __call__(self, *args, **kwargs):
        return self

class Mixins(
    ResultsExportMixin,
    ResultsFetchMixin,
    ResultsGGMixin,
):
    def long(self):
        return self.table().long()

    def print_long(self, max_rows: int = None) -> None:
        """Print the results in long format.

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling').print_long(max_rows = 2)
        ┏━━━━━━━━━━━━━━┳━━━━━━━━━━━━━┳━━━━━━━┓
        ┃ Result index ┃ Key         ┃ Value ┃
        ┡━━━━━━━━━━━━━━╇━━━━━━━━━━━━━╇━━━━━━━┩
        │ 0            │ how_feeling │ OK    │
        │ 1            │ how_feeling │ Great │
        └──────────────┴─────────────┴───────┘
        """
        from edsl.utilities.interface import print_results_long

        print_results_long(self, max_rows=max_rows)


class Results(UserList, Mixins, Base):
    """
    This class is a UserList of Result objects.

    It is instantiated with a `Survey` and a list of `Result` objects.
    It can be manipulated in various ways with select, filter, mutate, etc.
    It also has a list of created_columns, which are columns that have been created with `mutate` and are not part of the original data.
    """

    __documentation__ = "https://docs.expectedparrot.com/en/latest/results.html"

    known_data_types = [
        "answer",
        "scenario",
        "agent",
        "model",
        "prompt",
        "raw_model_response",
        "iteration",
        "question_text",
        "question_options",
        "question_type",
        "comment",
        "generated_tokens",
        "cache_used",
        "cache_keys",
    ]

    @classmethod
    def from_job_info(cls, job_info: dict) -> Results:
        """
        Instantiate a `Results` object from a job info dictionary.
        """
        results = cls()
        results.completed = False
        results.job_info = job_info
        return results

    def __init__(
        self,
        survey: Optional[Survey] = None,
        data: Optional[list[Result]] = None,
        created_columns: Optional[list[str]] = None,
        cache: Optional[Cache] = None,
        job_uuid: Optional[str] = None,
        total_results: Optional[int] = None,
        task_history: Optional[TaskHistory] = None,
    ):
        """Instantiate a `Results` object with a survey and a list of `Result` objects.

        :param survey: A Survey object.
        :param data: A list of Result objects.
        :param created_columns: A list of strings that are created columns.
        :param job_uuid: A string representing the job UUID.
        :param total_results: An integer representing the total number of results.
        :cache: A Cache object.
        """
        self.completed = True
        self._fetching = False
        super().__init__(data)
        from edsl.data.Cache import Cache
        from edsl.jobs.tasks.TaskHistory import TaskHistory

        self.survey = survey
        self.created_columns = created_columns or []
        self._job_uuid = job_uuid
        self._total_results = total_results
        self.cache = cache or Cache()

        self.task_history = task_history or TaskHistory(interviews=[])

        if hasattr(self, "_add_output_functions"):
            self._add_output_functions()

    def _summary(self) -> dict:
        import reprlib

        d = {
            "observations": len(self),
            "agents": len(set(self.agents)),
            "models": len(set(self.models)),
            "scenarios": len(set(self.scenarios)),
            "questions": len(self.survey),
            "Survey question names": reprlib.repr(self.survey.question_names),
        }
        return d

    def _cache_keys(self):
        cache_keys = []
        for result in self:
            cache_keys.extend(list(result["cache_keys"].values()))
        return cache_keys

    def relevant_cache(self, cache: Cache) -> Cache:
        cache_keys = self._cache_keys()
        return cache.subset(cache_keys)

    def insert(self, item):
        item_order = getattr(item, "order", None)
        if item_order is not None:
            # Get list of orders, putting None at the end
            orders = [getattr(x, "order", None) for x in self]
            # Filter to just the non-None orders for bisect
            sorted_orders = [x for x in orders if x is not None]
            if sorted_orders:
                index = bisect_left(sorted_orders, item_order)
                # Account for any None values before this position
                index += orders[:index].count(None)
            else:
                # If no sorted items yet, insert before any unordered items
                index = 0
            self.data.insert(index, item)
        else:
            # No order - append to end
            self.data.append(item)

    def append(self, item):
        self.insert(item)

    def extend(self, other):
        for item in other:
            self.insert(item)

    def compute_job_cost(self, include_cached_responses_in_cost: bool = False) -> float:
        """
        Computes the cost of a completed job in USD.
        """
        total_cost = 0
        for result in self:
            for key in result["raw_model_response"]:
                if key.endswith("_cost"):
                    result_cost = result["raw_model_response"][key]

                    question_name = key.removesuffix("_cost")
                    cache_used = result["cache_used_dict"][question_name]

                    if isinstance(result_cost, (int, float)):
                        if include_cached_responses_in_cost:
                            total_cost += result_cost
                        elif not include_cached_responses_in_cost and not cache_used:
                            total_cost += result_cost

        return total_cost

    def leaves(self):
        leaves = []
        for result in self:
            leaves.extend(result.leaves())
        return leaves

    def tree(self, node_list: Optional[List[str]] = None):
        return self.to_scenario_list().tree(node_list)

    def interactive_tree(
        self,
        fold_attributes: Optional[List[str]] = None,
        drop: Optional[List[str]] = None,
        open_file=True,
    ) -> dict:
        """Return the results as a tree."""
        from edsl.results.tree_explore import FoldableHTMLTableGenerator

        if drop is None:
            drop = []

        valid_attributes = [
            "model",
            "scenario",
            "agent",
            "answer",
            "question",
            "iteration",
        ]
        if fold_attributes is None:
            fold_attributes = []

        for attribute in fold_attributes:
            if attribute not in valid_attributes:
                raise ValueError(
                    f"Invalid fold attribute: {attribute}; must be in {valid_attributes}"
                )
        data = self.leaves()
        generator = FoldableHTMLTableGenerator(data)
        tree = generator.tree(fold_attributes=fold_attributes, drop=drop)
        html_content = generator.generate_html(tree, fold_attributes)
        import tempfile
        from edsl.utilities.utilities import is_notebook

        from IPython.display import display, HTML

        if is_notebook():
            import html
            from IPython.display import display, HTML

            height = 1000
            width = 1000
            escaped_output = html.escape(html_content)
            # escaped_output = rendered_html
            iframe = f""""
            <iframe srcdoc="{ escaped_output }" style="width: {width}px; height: {height}px;"></iframe>
            """
            display(HTML(iframe))
            return None

        with tempfile.NamedTemporaryFile(suffix=".html", delete=False) as f:
            f.write(html_content.encode())
            print(f"HTML file has been generated: {f.name}")

            if open_file:
                import webbrowser
                import time

                time.sleep(1)  # Wait for 1 second
                # webbrowser.open(f.name)
                import os

                filename = f.name
                webbrowser.open(f"file://{os.path.abspath(filename)}")

            else:
                return html_content

    def code(self):
        raise NotImplementedError

    def __getitem__(self, i):
        if isinstance(i, int):
            return self.data[i]

        if isinstance(i, slice):
            return self.__class__(survey=self.survey, data=self.data[i])

        if isinstance(i, str):
            return self.to_dict()[i]

        raise TypeError("Invalid argument type")

    def __add__(self, other: Results) -> Results:
        """Add two Results objects together.
        They must have the same survey and created columns.
        :param other: A Results object.

        Example:

        >>> r = Results.example()
        >>> r2 = Results.example()
        >>> r3 = r + r2
        """
        if self.survey != other.survey:
            raise ResultsError(
                "The surveys are not the same so the the results cannot be added together."
            )
        if self.created_columns != other.created_columns:
            raise ResultsError(
                "The created columns are not the same so they cannot be added together."
            )

        return Results(
            survey=self.survey,
            data=self.data + other.data,
            created_columns=self.created_columns,
        )
    
    def _repr_html_(self):
        if not self.completed:
            if hasattr(self, "job_info"):
                self.fetch_remote(self.job_info)
            
            if not self.completed:
                return f"Results not ready to call"
        
        return super()._repr_html_()

    # @ensure_ready
    # def __str__(self):
    #     super().__str__()    
    
    @ensure_ready
    def __repr__(self) -> str:
        return f"Results(data = {self.data}, survey = {repr(self.survey)}, created_columns = {self.created_columns})"

    def table(
        self,
        *fields,
        tablefmt: Optional[str] = None,
        pretty_labels: Optional[dict] = None,
        print_parameters: Optional[dict] = None,
    ):
        new_fields = []
        for field in fields:
            if "." in field:
                data_type, key = field.split(".")
                if data_type not in self.known_data_types:
                    raise ResultsInvalidNameError(
                        f"{data_type} is not a valid data type. Must be in {self.known_data_types}"
                    )
                if key == "*":
                    for k in self._data_type_to_keys[data_type]:
                        new_fields.append(k)
                else:
                    if key not in self._key_to_data_type:
                        raise ResultsColumnNotFoundError(
                            f"{key} is not a valid key. Must be in {self._key_to_data_type}"
                        )
                    new_fields.append(key)
            else:
                new_fields.append(field)

        return (
            self.to_scenario_list()
            .to_dataset()
            .table(
                *new_fields,
                tablefmt=tablefmt,
                pretty_labels=pretty_labels,
                print_parameters=print_parameters,
            )
        )

    def to_dict(
        self,
        sort: bool = False,
        add_edsl_version: bool = False,
        include_cache: bool = True,
        include_task_history: bool = False,
        include_cache_info: bool = True,
    ) -> dict[str, Any]:
        from edsl.data.Cache import Cache

        if sort:
            data = sorted([result for result in self.data], key=lambda x: hash(x))
        else:
            data = [result for result in self.data]

        d = {
            "data": [
                result.to_dict(
                    add_edsl_version=add_edsl_version,
                    include_cache_info=include_cache_info,
                )
                for result in data
            ],
            "survey": self.survey.to_dict(add_edsl_version=add_edsl_version),
            "created_columns": self.created_columns,
        }
        if include_cache:
            d.update(
                {
                    "cache": (
                        Cache()
                        if not hasattr(self, "cache")
                        else self.cache.to_dict(add_edsl_version=add_edsl_version)
                    )
                }
            )

        if self.task_history.has_unfixed_exceptions or include_task_history:
            d.update({"task_history": self.task_history.to_dict()})

        if add_edsl_version:
            from edsl import __version__

            d["edsl_version"] = __version__
            d["edsl_class_name"] = "Results"

        return d

    def compare(self, other_results: Results) -> dict:
        """
        Compare two Results objects and return the differences.
        """
        hashes_0 = [hash(result) for result in self]
        hashes_1 = [hash(result) for result in other_results]

        in_self_but_not_other = set(hashes_0).difference(set(hashes_1))
        in_other_but_not_self = set(hashes_1).difference(set(hashes_0))

        indicies_self = [hashes_0.index(h) for h in in_self_but_not_other]
        indices_other = [hashes_1.index(h) for h in in_other_but_not_self]
        return {
            "a_not_b": [self[i] for i in indicies_self],
            "b_not_a": [other_results[i] for i in indices_other],
        }

    @property
    def has_unfixed_exceptions(self) -> bool:
        return self.task_history.has_unfixed_exceptions

    def __hash__(self) -> int:
        from edsl.utilities.utilities import dict_hash

        return dict_hash(
            self.to_dict(sort=True, add_edsl_version=False, include_cache_info=False)
        )

    @property
    def hashes(self) -> set:
        return set(hash(result) for result in self.data)

    def sample(self, n: int) -> Results:
        """Return a random sample of the results.

        :param n: The number of samples to return.

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> len(r.sample(2))
        2
        """
        indices = None

        for entry in self:
            key, values = list(entry.items())[0]
            if indices is None:  # gets the indices for the first time
                indices = list(range(len(values)))
                sampled_indices = random.sample(indices, n)
                if n > len(indices):
                    raise ResultsError(
                        f"Cannot sample {n} items from a list of length {len(indices)}."
                    )
            entry[key] = [values[i] for i in sampled_indices]

        return self

    @classmethod
    @remove_edsl_version
    def from_dict(cls, data: dict[str, Any]) -> Results:
        """Convert a dictionary to a Results object.

        :param data: A dictionary representation of a Results object.

        Example:

        >>> r = Results.example()
        >>> d = r.to_dict()
        >>> r2 = Results.from_dict(d)
        >>> r == r2
        True
        """
        from edsl.surveys.Survey import Survey
        from edsl.data.Cache import Cache
        from edsl.results.Result import Result
        from edsl.jobs.tasks.TaskHistory import TaskHistory
        from edsl.agents.Agent import Agent

        survey = Survey.from_dict(data["survey"])
        results_data = [Result.from_dict(r) for r in data["data"]]
        created_columns = data.get("created_columns", None)
        cache = Cache.from_dict(data.get("cache")) if "cache" in data else Cache()
        task_history = (
            TaskHistory.from_dict(data.get("task_history"))
            if "task_history" in data
            else TaskHistory(interviews=[])
        )
        params = {
            "survey": survey,
            "data": results_data,
            "created_columns": created_columns,
            "cache": cache,
            "task_history": task_history,
        }

        try:
            results = cls(**params)
        except Exception as e:
            raise ResultsDeserializationError(f"Error in Results.from_dict: {e}")
        return results

    @property
    def _key_to_data_type(self) -> dict[str, str]:
        """
        Return a mapping of keys (how_feeling, status, etc.) to strings representing data types.

        Objects such as Agent, Answer, Model, Scenario, etc.
        - Uses the key_to_data_type property of the Result class.
        - Includes any columns that the user has created with `mutate`
        """
        d: dict = {}
        for result in self.data:
            d.update(result.key_to_data_type)
        for column in self.created_columns:
            d[column] = "answer"

        return d

    @property
    def _data_type_to_keys(self) -> dict[str, str]:
        """
        Return a mapping of strings representing data types (objects such as Agent, Answer, Model, Scenario, etc.) to keys (how_feeling, status, etc.)
        - Uses the key_to_data_type property of the Result class.
        - Includes any columns that the user has created with `mutate`

        Example:

        >>> r = Results.example()
        >>> r._data_type_to_keys
        defaultdict(...
        """
        d: dict = defaultdict(set)
        for result in self.data:
            for key, value in result.key_to_data_type.items():
                d[value] = d[value].union(set({key}))
        for column in self.created_columns:
            d["answer"] = d["answer"].union(set({column}))
        return d

    @property
    def columns(self) -> list[str]:
        """Return a list of all of the columns that are in the Results.

        Example:

        >>> r = Results.example()
        >>> r.columns
        ['agent.agent_index', ...]
        """
        column_names = [f"{v}.{k}" for k, v in self._key_to_data_type.items()]
        from edsl.utilities.PrettyList import PrettyList

        return PrettyList(sorted(column_names))

    @property
    def answer_keys(self) -> dict[str, str]:
        """Return a mapping of answer keys to question text.

        Example:

        >>> r = Results.example()
        >>> r.answer_keys
        {'how_feeling': 'How are you this {{ period }}?', 'how_feeling_yesterday': 'How were you feeling yesterday {{ period }}?'}
        """
        from edsl.utilities.utilities import shorten_string

        if not self.survey:
            raise ResultsError("Survey is not defined so no answer keys are available.")

        answer_keys = self._data_type_to_keys["answer"]
        answer_keys = {k for k in answer_keys if "_comment" not in k}
        questions_text = [
            self.survey._get_question_by_name(k).question_text for k in answer_keys
        ]
        short_question_text = [shorten_string(q, 80) for q in questions_text]
        initial_dict = dict(zip(answer_keys, short_question_text))
        sorted_dict = {key: initial_dict[key] for key in sorted(initial_dict)}
        return sorted_dict

    @property
    def agents(self) -> AgentList:
        """Return a list of all of the agents in the Results.

        Example:

        >>> r = Results.example()
        >>> r.agents
        AgentList([Agent(traits = {'status': 'Joyful'}), Agent(traits = {'status': 'Joyful'}), Agent(traits = {'status': 'Sad'}), Agent(traits = {'status': 'Sad'})])
        """
        from edsl.agents.AgentList import AgentList

        return AgentList([r.agent for r in self.data])

    @property
    def models(self) -> ModelList:
        """Return a list of all of the models in the Results.

        Example:

        >>> r = Results.example()
        >>> r.models[0]
        Model(model_name = ...)
        """
        from edsl.language_models.ModelList import ModelList

        return ModelList([r.model for r in self.data])

    def __eq__(self, other):
        return hash(self) == hash(other)

    @property
    def scenarios(self) -> ScenarioList:
        """Return a list of all of the scenarios in the Results.

        Example:

        >>> r = Results.example()
        >>> r.scenarios
        ScenarioList([Scenario({'period': 'morning', 'scenario_index': 0}), Scenario({'period': 'afternoon', 'scenario_index': 1}), Scenario({'period': 'morning', 'scenario_index': 0}), Scenario({'period': 'afternoon', 'scenario_index': 1})])
        """
        from edsl.scenarios.ScenarioList import ScenarioList

        return ScenarioList([r.scenario for r in self.data])

    @property
    def agent_keys(self) -> list[str]:
        """Return a set of all of the keys that are in the Agent data.

        Example:

        >>> r = Results.example()
        >>> r.agent_keys
        ['agent_index', 'agent_instruction', 'agent_name', 'status']
        """
        return sorted(self._data_type_to_keys["agent"])

    @property
    def model_keys(self) -> list[str]:
        """Return a set of all of the keys that are in the LanguageModel data.

        >>> r = Results.example()
        >>> r.model_keys
        ['frequency_penalty', 'inference_service', 'logprobs', 'max_tokens', 'model', 'model_index', 'presence_penalty', 'temperature', 'top_logprobs', 'top_p']
        """
        return sorted(self._data_type_to_keys["model"])

    @property
    def scenario_keys(self) -> list[str]:
        """Return a set of all of the keys that are in the Scenario data.

        >>> r = Results.example()
        >>> r.scenario_keys
        ['period', 'scenario_index']
        """
        return sorted(self._data_type_to_keys["scenario"])

    @property
    def question_names(self) -> list[str]:
        """Return a list of all of the question names.

        Example:

        >>> r = Results.example()
        >>> r.question_names
        ['how_feeling', 'how_feeling_yesterday']
        """
        if self.survey is None:
            return []
        return sorted(list(self.survey.question_names))

    @property
    def all_keys(self) -> list[str]:
        """Return a set of all of the keys that are in the Results.

        Example:

        >>> r = Results.example()
        >>> r.all_keys
        ['agent_index', ...]
        """
        answer_keys = set(self.answer_keys)
        all_keys = (
            answer_keys.union(self.agent_keys)
            .union(self.scenario_keys)
            .union(self.model_keys)
        )
        return sorted(list(all_keys))

    def first(self) -> Result:
        """Return the first observation in the results.

        Example:

        >>> r = Results.example()
        >>> r.first()
        Result(agent...
        """
        return self.data[0]

    def answer_truncate(
        self, column: str, top_n: int = 5, new_var_name: str = None
    ) -> Results:
        """Create a new variable that truncates the answers to the top_n.

        :param column: The column to truncate.
        :param top_n: The number of top answers to keep.
        :param new_var_name: The name of the new variable. If None, it is the original name + '_truncated'.

        Example:
        >>> r = Results.example()
        >>> r.answer_truncate('how_feeling', top_n = 2).select('how_feeling', 'how_feeling_truncated')
        Dataset([{'answer.how_feeling': ['OK', 'Great', 'Terrible', 'OK']}, {'answer.how_feeling_truncated': ['Other', 'Other', 'Other', 'Other']}])


        """
        if new_var_name is None:
            new_var_name = column + "_truncated"
        answers = list(self.select(column).tally().keys())

        def f(x):
            if x in answers[:top_n]:
                return x
            else:
                return "Other"

        return self.recode(column, recode_function=f, new_var_name=new_var_name)

    @ensure_ready
    def recode(
        self, column: str, recode_function: Optional[Callable], new_var_name=None
    ) -> Results:
        """
        Recode a column in the Results object.

        >>> r = Results.example()
        >>> r.recode('how_feeling', recode_function = lambda x: 1 if x == 'Great' else 0).select('how_feeling', 'how_feeling_recoded')
        Dataset([{'answer.how_feeling': ['OK', 'Great', 'Terrible', 'OK']}, {'answer.how_feeling_recoded': [0, 1, 0, 0]}])
        """

        if new_var_name is None:
            new_var_name = column + "_recoded"
        new_data = []
        for result in self.data:
            new_result = result.copy()
            value = new_result.get_value("answer", column)
            # breakpoint()
            new_result["answer"][new_var_name] = recode_function(value)
            new_data.append(new_result)

        # print("Created new variable", new_var_name)
        return Results(
            survey=self.survey,
            data=new_data,
            created_columns=self.created_columns + [new_var_name],
        )

    @ensure_ready
    def add_column(self, column_name: str, values: list) -> Results:
        """Adds columns to Results

        >>> r = Results.example()
        >>> r.add_column('a', [1,2,3, 4]).select('a')
        Dataset([{'answer.a': [1, 2, 3, 4]}])
        """

        assert len(values) == len(
            self.data
        ), "The number of values must match the number of results."
        new_results = self.data.copy()
        for i, result in enumerate(new_results):
            result["answer"][column_name] = values[i]
        return Results(
            survey=self.survey,
            data=new_results,
            created_columns=self.created_columns + [column_name],
        )

    @ensure_ready
    def add_columns_from_dict(self, columns: List[dict]) -> Results:
        """Adds columns to Results from a list of dictionaries.

        >>> r = Results.example()
        >>> r.add_columns_from_dict([{'a': 1, 'b': 2}, {'a': 3, 'b': 4}, {'a':3, 'b':2}, {'a':3, 'b':2}]).select('a', 'b')
        Dataset([{'answer.a': [1, 3, 3, 3]}, {'answer.b': [2, 4, 2, 2]}])
        """
        keys = list(columns[0].keys())
        for key in keys:
            values = [d[key] for d in columns]
            self = self.add_column(key, values)
        return self

    @staticmethod
    def _create_evaluator(
        result: Result, functions_dict: Optional[dict] = None
    ) -> "EvalWithCompoundTypes":
        """Create an evaluator for the expression.

        >>> from unittest.mock import Mock
        >>> result = Mock()
        >>> result.combined_dict = {'how_feeling': 'OK'}

        >>> evaluator = Results._create_evaluator(result = result, functions_dict = {})
        >>> evaluator.eval("how_feeling == 'OK'")
        True

        >>> result.combined_dict = {'answer': {'how_feeling': 'OK'}}
        >>> evaluator = Results._create_evaluator(result = result, functions_dict = {})
        >>> evaluator.eval("answer.how_feeling== 'OK'")
        True

        Note that you need to refer to the answer dictionary in the expression.

        >>> evaluator.eval("how_feeling== 'OK'")
        Traceback (most recent call last):
        ...
        simpleeval.NameNotDefined: 'how_feeling' is not defined for expression 'how_feeling== 'OK''
        """
        from simpleeval import EvalWithCompoundTypes

        if functions_dict is None:
            functions_dict = {}
        evaluator = EvalWithCompoundTypes(
            names=result.combined_dict, functions=functions_dict
        )
        evaluator.functions.update(int=int, float=float)
        return evaluator

    @ensure_ready
    def mutate(
        self, new_var_string: str, functions_dict: Optional[dict] = None
    ) -> Results:
        """
        Creates a value in the Results object as if has been asked as part of the survey.

        :param new_var_string: A string that is a valid Python expression.
        :param functions_dict: A dictionary of functions that can be used in the expression. The keys are the function names and the values are the functions themselves.

        It splits the new_var_string at the "=" and uses simple_eval

        Example:

        >>> r = Results.example()
        >>> r.mutate('how_feeling_x = how_feeling + "x"').select('how_feeling_x')
        Dataset([{'answer.how_feeling_x': ...
        """
        # extract the variable name and the expression
        if "=" not in new_var_string:
            raise ResultsBadMutationstringError(
                f"Mutate requires an '=' in the string, but '{new_var_string}' doesn't have one."
            )
        raw_var_name, expression = new_var_string.split("=", 1)
        var_name = raw_var_name.strip()
        from edsl.utilities.utilities import is_valid_variable_name

        if not is_valid_variable_name(var_name):
            raise ResultsInvalidNameError(f"{var_name} is not a valid variable name.")

        # create the evaluator
        functions_dict = functions_dict or {}

        def new_result(old_result: "Result", var_name: str) -> "Result":
            evaluator = self._create_evaluator(old_result, functions_dict)
            value = evaluator.eval(expression)
            new_result = old_result.copy()
            new_result["answer"][var_name] = value
            return new_result

        try:
            new_data = [new_result(result, var_name) for result in self.data]
        except Exception as e:
            raise ResultsMutateError(f"Error in mutate. Exception:{e}")

        return Results(
            survey=self.survey,
            data=new_data,
            created_columns=self.created_columns + [var_name],
        )

    @ensure_ready
    def add_column(self, column_name: str, values: list) -> Results:
        """Adds columns to Results

        >>> r = Results.example()
        >>> r.add_column('a', [1,2,3, 4]).select('a')
        Dataset([{'answer.a': [1, 2, 3, 4]}])
        """

        assert len(values) == len(
            self.data
        ), "The number of values must match the number of results."
        new_results = self.data.copy()
        for i, result in enumerate(new_results):
            result["answer"][column_name] = values[i]
        return Results(
            survey=self.survey,
            data=new_results,
            created_columns=self.created_columns + [column_name],
        )

    @ensure_ready
    def rename(self, old_name: str, new_name: str) -> Results:
        """Rename an answer column in a Results object.

        >>> s = Results.example()
        >>> s.rename('how_feeling', 'how_feeling_new').select('how_feeling_new')
        Dataset([{'answer.how_feeling_new': ['OK', 'Great', 'Terrible', 'OK']}])

        # TODO: Should we allow renaming of scenario fields as well? Probably.

        """

        for obs in self.data:
            obs["answer"][new_name] = obs["answer"][old_name]
            del obs["answer"][old_name]

        return self

    @ensure_ready
    def shuffle(self, seed: Optional[str] = "edsl") -> Results:
        """Shuffle the results.

        Example:

        >>> r = Results.example()
        >>> r.shuffle(seed = 1)[0]
        Result(...)
        """
        if seed != "edsl":
            seed = random.seed(seed)

        new_data = self.data.copy()
        random.shuffle(new_data)
        return Results(survey=self.survey, data=new_data, created_columns=None)

    @ensure_ready
    def sample(
        self,
        n: Optional[int] = None,
        frac: Optional[float] = None,
        with_replacement: bool = True,
        seed: Optional[str] = None,
    ) -> Results:
        """Sample the results.

        :param n: An integer representing the number of samples to take.
        :param frac: A float representing the fraction of samples to take.
        :param with_replacement: A boolean representing whether to sample with replacement.
        :param seed: An integer representing the seed for the random number generator.

        Example:

        >>> r = Results.example()
        >>> len(r.sample(2))
        2
        """
        if seed:
            random.seed(seed)

        if n is None and frac is None:
            raise Exception("You must specify either n or frac.")

        if n is not None and frac is not None:
            raise Exception("You cannot specify both n and frac.")

        if frac is not None and n is None:
            n = int(frac * len(self.data))

        if with_replacement:
            new_data = random.choices(self.data, k=n)
        else:
            new_data = random.sample(self.data, n)

        return Results(survey=self.survey, data=new_data, created_columns=None)

    @ensure_ready
    def select(self, *columns: Union[str, list[str]]) -> Results:
        """
        Select data from the results and format it.

        :param columns: A list of strings, each of which is a column name. The column name can be a single key, e.g. "how_feeling", or a dot-separated string, e.g. "answer.how_feeling".

        Example:

        >>> results = Results.example()
        >>> results.select('how_feeling')
        Dataset([{'answer.how_feeling': ['OK', 'Great', 'Terrible', 'OK']}])

        >>> results.select('how_feeling', 'model', 'how_feeling')
        Dataset([{'answer.how_feeling': ['OK', 'Great', 'Terrible', 'OK']}, {'answer.how_feeling': ['OK', 'Great', 'Terrible', 'OK']}, {'model.model': ['...', '...', '...', '...']}, {'answer.how_feeling': ['OK', 'Great', 'Terrible', 'OK']}, {'answer.how_feeling': ['OK', 'Great', 'Terrible', 'OK']}])

        >>> from edsl import Results; r = Results.example(); r.select('answer.how_feeling_y')
        Dataset([{'answer.how_feeling_yesterday': ['Great', 'Good', 'OK', 'Terrible']}])
        """

        from edsl.results.results_selector import Selector

        if len(self) == 0:
            raise Exception("No data to select from---the Results object is empty.")

        selector = Selector(
            known_data_types=self.known_data_types,
            data_type_to_keys=self._data_type_to_keys,
            key_to_data_type=self._key_to_data_type,
            fetch_list_func=self._fetch_list,
            columns=self.columns,
        )
        return selector.select(*columns)

    @ensure_ready
    def sort_by(self, *columns: str, reverse: bool = False) -> Results:
        """Sort the results by one or more columns."""
        import warnings

        warnings.warn(
            "sort_by is deprecated. Use order_by instead.", DeprecationWarning
        )
        return self.order_by(*columns, reverse=reverse)

    def _parse_column(self, column: str) -> tuple[str, str]:
        """Parse a column name into a data type and key."""
        if "." in column:
            return column.split(".")
        return self._key_to_data_type[column], column

    @ensure_ready
    def order_by(self, *columns: str, reverse: bool = False) -> Results:
        """Sort the results by one or more columns.

        :param columns: One or more column names as strings.
        :param reverse: A boolean that determines whether to sort in reverse order.

        Each column name can be a single key, e.g. "how_feeling", or a dot-separated string, e.g. "answer.how_feeling".

        Example:

        >>> r = Results.example()
        >>> r.sort_by('how_feeling', reverse=False).select('how_feeling')
        Dataset([{'answer.how_feeling': ['Great', 'OK', 'OK', 'Terrible']}])

        >>> r.sort_by('how_feeling', reverse=True).select('how_feeling')
        Dataset([{'answer.how_feeling': ['Terrible', 'OK', 'OK', 'Great']}])

        """

        def to_numeric_if_possible(v):
            try:
                return float(v)
            except:
                return v

        def sort_key(item):
            key_components = []
            for col in columns:
                data_type, key = self._parse_column(col)
                value = item.get_value(data_type, key)
                key_components.append(to_numeric_if_possible(value))
            return tuple(key_components)

        new_data = sorted(self.data, key=sort_key, reverse=reverse)
        return Results(survey=self.survey, data=new_data, created_columns=None)

    @ensure_ready
    def filter(self, expression: str) -> Results:
        """
        Filter based on the given expression and returns the filtered `Results`.

        :param expression: A string expression that evaluates to a boolean. The expression is applied to each element in `Results` to determine whether it should be included in the filtered results.

        The `expression` parameter is a string that must resolve to a boolean value when evaluated against each element in `Results`.
        This expression is used to determine which elements to include in the returned `Results`.

        Example usage: Create an example `Results` instance and apply filters to it:

        >>> r = Results.example()
        >>> r.filter("how_feeling == 'Great'").select('how_feeling')
        Dataset([{'answer.how_feeling': ['Great']}])

        Example usage: Using an OR operator in the filter expression.

        >>> r = Results.example().filter("how_feeling = 'Great'").select('how_feeling')
        Traceback (most recent call last):
        ...
        edsl.exceptions.results.ResultsFilterError: You must use '==' instead of '=' in the filter expression.
        ...

        >>> r.filter("how_feeling == 'Great' or how_feeling == 'Terrible'").select('how_feeling')
        Dataset([{'answer.how_feeling': ['Great', 'Terrible']}])
        """

        def has_single_equals(string):
            if "!=" in string:
                return False
            if "=" in string and not (
                "==" in string or "<=" in string or ">=" in string
            ):
                return True

        if has_single_equals(expression):
            raise ResultsFilterError(
                "You must use '==' instead of '=' in the filter expression."
            )

        try:
            # iterates through all the results and evaluates the expression
            new_data = []
            for result in self.data:
                evaluator = self._create_evaluator(result)
                result.check_expression(expression)  # check expression
                if evaluator.eval(expression):
                    new_data.append(result)

        except ValueError as e:
            raise ResultsFilterError(
                f"Error in filter. Exception:{e}",
                f"The expression you provided was: {expression}",
                "See https://docs.expectedparrot.com/en/latest/results.html#filtering-results for more details.",
            )
        except Exception as e:
            raise ResultsFilterError(
                f"""Error in filter. Exception:{e}.""",
                f"""The expression you provided was: {expression}.""",
                """Please make sure that the expression is a valid Python expression that evaluates to a boolean.""",
                """For example, 'how_feeling == "Great"' is a valid expression, as is 'how_feeling in ["Great", "Terrible"]'., """,
                """However, 'how_feeling = "Great"' is not a valid expression.""",
                """See https://docs.expectedparrot.com/en/latest/results.html#filtering-results for more details.""",
            )

        if len(new_data) == 0:
            import warnings

            warnings.warn("No results remain after applying the filter.")

        return Results(survey=self.survey, data=new_data, created_columns=None)

    @classmethod
    def example(cls, randomize: bool = False) -> Results:
        """Return an example `Results` object.

        Example usage:

        >>> r = Results.example()

        :param debug: if False, uses actual API calls
        """
        from edsl.jobs.Jobs import Jobs
        from edsl.data.Cache import Cache

        c = Cache()
        job = Jobs.example(randomize=randomize)
        results = job.run(
            cache=c,
            stop_on_exception=True,
            skip_retry=True,
            raise_validation_errors=True,
            disable_remote_cache=True,
            disable_remote_inference=True,
        )
        return results

    def rich_print(self):
        """Display an object as a table."""
        pass

    @ensure_ready
    def __str__(self):
        data = self.to_dict()["data"]
        return json.dumps(data, indent=4)

    def show_exceptions(self, traceback=False):
        """Print the exceptions."""
        if hasattr(self, "task_history"):
            self.task_history.show_exceptions(traceback)
        else:
            print("No exceptions to show.")

    def score(self, f: Callable) -> list:
        """Score the results using in a function.

        :param f: A function that takes values from a Resul object and returns a score.

        >>> r = Results.example()
        >>> def f(status): return 1 if status == 'Joyful' else 0
        >>> r.score(f)
        [1, 1, 0, 0]
        """
        return [r.score(f) for r in self.data]
    

    def fetch_remote(self, job_info: "RemoteJobInfo") -> None:
        """
        Fetches the remote Results object using the provided RemoteJobInfo and updates this instance with the remote data.
        
        This is useful when you have a Results object that was created locally but want to sync it with 
        the latest data from the remote server.
        
        Args:
            job_info: RemoteJobInfo object containing the job_uuid and other remote job details
        
        """
        #print("Calling fetch_remote")
        try:
            from edsl.coop.coop import Coop
            from edsl.jobs.JobsRemoteInferenceHandler import JobsRemoteInferenceHandler
            
            # Get the remote job data
            remote_job_data = JobsRemoteInferenceHandler.check_status(job_info.job_uuid)
            
            if remote_job_data.get("status") not in ["completed", "failed"]:
                return False
                #            
            results_uuid = remote_job_data.get("results_uuid")
            if not results_uuid:
                raise ResultsError("No results_uuid found in remote job data")
            
            # Fetch the remote Results object
            coop = Coop()
            remote_results = coop.get(results_uuid, expected_object_type="results")
            
            # Update this instance with remote data
            self.data = remote_results.data
            self.survey = remote_results.survey
            self.created_columns = remote_results.created_columns
            self.cache = remote_results.cache
            self.task_history = remote_results.task_history
            self.completed = True
            
            # Set job_uuid and results_uuid from remote data
            self.job_uuid = job_info.job_uuid
            if hasattr(remote_results, 'results_uuid'):
                self.results_uuid = remote_results.results_uuid

            return True

        except Exception as e:
            raise ResultsError(f"Failed to fetch remote results: {str(e)}")

    def fetch(self, polling_interval: float = 1.0) -> Results:
        """
        Polls the server for job completion and updates this Results instance with the completed data.
        
        Args:
            polling_interval: Number of seconds to wait between polling attempts (default: 1.0)
        
        Returns:
            self: The updated Results instance
        """
        if not hasattr(self, "job_info"):
            raise ResultsError("No job info available - this Results object wasn't created from a remote job")
        
        from edsl.jobs.JobsRemoteInferenceHandler import JobsRemoteInferenceHandler
        
        try:
            # Get the remote job data
            remote_job_data = JobsRemoteInferenceHandler.check_status(self.job_info.job_uuid)
            
            while remote_job_data.get("status") not in ["completed", "failed"]:
                import time
                time.sleep(polling_interval)
                remote_job_data = JobsRemoteInferenceHandler.check_status(self.job_info.job_uuid)
                
            # Once complete, fetch the full results
            self.fetch_remote(self.job_info)
            return self
            
        except Exception as e:
            raise ResultsError(f"Failed to fetch remote results: {str(e)}")


def main():  # pragma: no cover
    """Call the OpenAI API credits."""
    from edsl.results.Results import Results

    results = Results.example(debug=True)
    print(results.filter("how_feeling == 'Great'").select("how_feeling"))
    print(results.mutate("how_feeling_x = how_feeling + 'x'").select("how_feeling_x"))


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



import os
import mimetypes


class ConstructDownloadLink:
    """
    A class to create HTML download links for FileStore objects.
    The links can be displayed in Jupyter notebooks or other web interfaces.
    """

    def __init__(self, filestore):
        """
        Initialize with a FileStore object.

        Args:
            filestore: A FileStore object containing the file to be made downloadable
        """
        self.filestore = filestore

    def create_link(self, custom_filename=None, style=None):
        from IPython.display import HTML

        html = self.html_create_link(custom_filename, style)
        return HTML(html)

    def html_create_link(self, custom_filename=None, style=None):
        """
        Create an HTML download link for the file.

        Args:
            custom_filename (str, optional): Custom name for the downloaded file.
                                          If None, uses original filename.
            style (dict, optional): Custom CSS styles for the download button.
                                  If None, uses default styling.

        Returns:
            IPython.display.HTML: HTML object containing the download link
        """

        # Get filename from path or use custom filename
        original_filename = os.path.basename(self.filestore.path)
        filename = custom_filename or original_filename

        # Use the base64 string already stored in FileStore
        b64_data = self.filestore.base64_string

        # Use mime type from FileStore or guess it
        mime_type = self.filestore.mime_type

        # Default style if none provided
        default_style = {
            "background-color": "#4CAF50",
            "color": "white",
            "padding": "10px 20px",
            "text-decoration": "none",
            "border-radius": "4px",
            "display": "inline-block",
            "margin": "10px 0",
            "font-family": "sans-serif",
            "cursor": "pointer",
        }

        button_style = style or default_style
        style_str = "; ".join(f"{k}: {v}" for k, v in button_style.items())

        html = f"""
        <a download="{filename}" 
           href="data:{mime_type};base64,{b64_data}" 
           style="{style_str}">
            Download {filename}
        </a>
        """
        return html

    def create_multiple_links(self, files, custom_filenames=None, style=None):
        """
        Create multiple download links at once.
        Useful when you want to provide different versions of the same file
        or related files together.

        Args:
            files (list): List of FileStore objects
            custom_filenames (list, optional): List of custom filenames for downloads
            style (dict, optional): Custom CSS styles for the download buttons

        Returns:
            IPython.display.HTML: HTML object containing all download links
        """
        if custom_filenames is None:
            custom_filenames = [None] * len(files)

        html_parts = []
        for file_obj, custom_name in zip(files, custom_filenames):
            link_creator = ConstructDownloadLink(file_obj)
            html_parts.append(
                link_creator.create_link(
                    custom_filename=custom_name, style=style
                )._repr_html_()
            )

        return HTML(
            '<div style="display: flex; gap: 10px;">' + "".join(html_parts) + "</div>"
        )


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import aiohttp
import json
import requests
from typing import Any, List, Optional

# from edsl.inference_services.InferenceServiceABC import InferenceServiceABC
from edsl.language_models import LanguageModel

from edsl.inference_services.OpenAIService import OpenAIService
import openai


class TogetherAIService(OpenAIService):
    """DeepInfra service class."""

    _inference_service_ = "together"
    _env_key_name_ = "TOGETHER_API_KEY"
    _base_url_ = "https://api.together.xyz/v1"
    _models_list_cache: List[str] = []

    # These are non-serverless models. There was no api param to filter them
    model_exclude_list = [
        "EleutherAI/llemma_7b",
        "HuggingFaceH4/zephyr-7b-beta",
        "Nexusflow/NexusRaven-V2-13B",
        "NousResearch/Hermes-2-Theta-Llama-3-70B",
        "NousResearch/Nous-Capybara-7B-V1p9",
        "NousResearch/Nous-Hermes-13b",
        "NousResearch/Nous-Hermes-2-Mistral-7B-DPO",
        "NousResearch/Nous-Hermes-2-Mixtral-8x7B-SFT",
        "NousResearch/Nous-Hermes-Llama2-13b",
        "NousResearch/Nous-Hermes-Llama2-70b",
        "NousResearch/Nous-Hermes-llama-2-7b",
        "NumbersStation/nsql-llama-2-7B",
        "Open-Orca/Mistral-7B-OpenOrca",
        "Phind/Phind-CodeLlama-34B-Python-v1",
        "Phind/Phind-CodeLlama-34B-v2",
        "Qwen/Qwen1.5-0.5B",
        "Qwen/Qwen1.5-0.5B-Chat",
        "Qwen/Qwen1.5-1.8B",
        "Qwen/Qwen1.5-1.8B-Chat",
        "Qwen/Qwen1.5-14B",
        "Qwen/Qwen1.5-14B-Chat",
        "Qwen/Qwen1.5-32B",
        "Qwen/Qwen1.5-32B-Chat",
        "Qwen/Qwen1.5-4B",
        "Qwen/Qwen1.5-4B-Chat",
        "Qwen/Qwen1.5-72B",
        "Qwen/Qwen1.5-7B",
        "Qwen/Qwen1.5-7B-Chat",
        "Qwen/Qwen2-1.5B",
        "Qwen/Qwen2-1.5B-Instruct",
        "Qwen/Qwen2-72B",
        "Qwen/Qwen2-7B",
        "Qwen/Qwen2-7B-Instruct",
        "SG161222/Realistic_Vision_V3.0_VAE",
        "Snowflake/snowflake-arctic-instruct",
        "Undi95/ReMM-SLERP-L2-13B",
        "Undi95/Toppy-M-7B",
        "WizardLM/WizardCoder-Python-34B-V1.0",
        "WizardLM/WizardLM-13B-V1.2",
        "WizardLM/WizardLM-70B-V1.0",
        "allenai/OLMo-7B",
        "allenai/OLMo-7B-Instruct",
        "bert-base-uncased",
        "codellama/CodeLlama-13b-Instruct-hf",
        "codellama/CodeLlama-13b-Python-hf",
        "codellama/CodeLlama-13b-hf",
        "codellama/CodeLlama-34b-Python-hf",
        "codellama/CodeLlama-34b-hf",
        "codellama/CodeLlama-70b-Instruct-hf",
        "codellama/CodeLlama-70b-Python-hf",
        "codellama/CodeLlama-70b-hf",
        "codellama/CodeLlama-7b-Instruct-hf",
        "codellama/CodeLlama-7b-Python-hf",
        "codellama/CodeLlama-7b-hf",
        "cognitivecomputations/dolphin-2.5-mixtral-8x7b",
        "deepseek-ai/deepseek-coder-33b-instruct",
        "garage-bAInd/Platypus2-70B-instruct",
        "google/gemma-2b",
        "google/gemma-7b",
        "google/gemma-7b-it",
        "gradientai/Llama-3-70B-Instruct-Gradient-1048k",
        "hazyresearch/M2-BERT-2k-Retrieval-Encoder-V1",
        "huggyllama/llama-13b",
        "huggyllama/llama-30b",
        "huggyllama/llama-65b",
        "huggyllama/llama-7b",
        "lmsys/vicuna-13b-v1.3",
        "lmsys/vicuna-13b-v1.5",
        "lmsys/vicuna-13b-v1.5-16k",
        "lmsys/vicuna-7b-v1.3",
        "lmsys/vicuna-7b-v1.5",
        "meta-llama/Llama-2-13b-hf",
        "meta-llama/Llama-2-70b-chat-hf",
        "meta-llama/Llama-2-7b-hf",
        "meta-llama/Llama-3-70b-hf",
        "meta-llama/Llama-3-8b-hf",
        "meta-llama/Meta-Llama-3-70B",
        "meta-llama/Meta-Llama-3-70B-Instruct",
        "meta-llama/Meta-Llama-3-8B-Instruct",
        "meta-llama/Meta-Llama-3.1-70B-Instruct-Reference",
        "meta-llama/Meta-Llama-3.1-70B-Reference",
        "meta-llama/Meta-Llama-3.1-8B-Reference",
        "microsoft/phi-2",
        "mistralai/Mixtral-8x22B",
        "openchat/openchat-3.5-1210",
        "prompthero/openjourney",
        "runwayml/stable-diffusion-v1-5",
        "sentence-transformers/msmarco-bert-base-dot-v5",
        "snorkelai/Snorkel-Mistral-PairRM-DPO",
        "stabilityai/stable-diffusion-2-1",
        "teknium/OpenHermes-2-Mistral-7B",
        "teknium/OpenHermes-2p5-Mistral-7B",
        "togethercomputer/CodeLlama-13b-Instruct",
        "togethercomputer/CodeLlama-13b-Python",
        "togethercomputer/CodeLlama-34b",
        "togethercomputer/CodeLlama-34b-Python",
        "togethercomputer/CodeLlama-7b-Instruct",
        "togethercomputer/CodeLlama-7b-Python",
        "togethercomputer/Koala-13B",
        "togethercomputer/Koala-7B",
        "togethercomputer/LLaMA-2-7B-32K",
        "togethercomputer/SOLAR-10.7B-Instruct-v1.0-int4",
        "togethercomputer/StripedHyena-Hessian-7B",
        "togethercomputer/alpaca-7b",
        "togethercomputer/evo-1-131k-base",
        "togethercomputer/evo-1-8k-base",
        "togethercomputer/guanaco-13b",
        "togethercomputer/guanaco-33b",
        "togethercomputer/guanaco-65b",
        "togethercomputer/guanaco-7b",
        "togethercomputer/llama-2-13b",
        "togethercomputer/llama-2-70b-chat",
        "togethercomputer/llama-2-7b",
        "wavymulder/Analog-Diffusion",
        "zero-one-ai/Yi-34B",
        "zero-one-ai/Yi-34B-Chat",
        "zero-one-ai/Yi-6B",
    ]

    _sync_client_ = openai.OpenAI
    _async_client_ = openai.AsyncOpenAI

    @classmethod
    def get_model_list(cls, api_token=None):
        # Togheter.ai has a different response in model list then openai
        # and the OpenAI class returns an error when calling .models.list()
        import requests
        import os

        url = "https://api.together.xyz/v1/models?filter=serverless"
        if api_token is None:
            api_token = os.getenv(cls._env_key_name_)

        headers = {"accept": "application/json", "authorization": f"Bearer {api_token}"}

        response = requests.get(url, headers=headers)
        return response.json()

    @classmethod
    def available(cls) -> List[str]:
        if not cls._models_list_cache:
            try:
                cls._models_list_cache = [
                    m["id"]
                    for m in cls.get_model_list()
                    if m["id"] not in cls.model_exclude_list
                ]
            except Exception as e:
                raise
        return cls._models_list_cache



class ConfigurationError(Exception):
    """Base exception for errors."""

    pass


class InvalidEnvironmentVariableError(ConfigurationError):
    """Raised when an environment variable is invalid."""

    pass


class MissingEnvironmentVariableError(ConfigurationError):
    """Raised when an expected environment variable is missing."""

    pass



from abc import ABC, abstractmethod
import io
import csv
import base64
from typing import Optional, Union, Tuple, List, Any, Dict
from openpyxl import Workbook

from edsl.scenarios.FileStore import FileStore


class FileExport(ABC):
    def __init__(
        self,
        data: Any,
        filename: Optional[str] = None,
        remove_prefix: bool = False,
        pretty_labels: Optional[Dict[str, str]] = None,
    ):
        self.data = data
        self.filename = filename  # or self._get_default_filename()
        self.remove_prefix = remove_prefix
        self.pretty_labels = pretty_labels

    @property
    def mime_type(self) -> str:
        """Return the MIME type for this export format."""
        return self.__class__.mime_type

    @property
    def suffix(self) -> str:
        """Return the file suffix for this format."""
        return self.__class__.suffix

    @property
    def is_binary(self) -> bool:
        """Whether the format is binary or text-based."""
        return self.__class__.is_binary

    def _get_default_filename(self) -> str:
        """Generate default filename for this format."""
        return f"results.{self.suffix}"

    def _create_filestore(self, data: Union[str, bytes]) -> "FileStore":
        """Create a FileStore instance with encoded data."""
        if isinstance(data, str):
            base64_string = base64.b64encode(data.encode()).decode()
        else:
            base64_string = base64.b64encode(data).decode()

        from edsl.scenarios.FileStore import FileStore

        path = self.filename or self._get_default_filename()

        fs = FileStore(
            path=path,
            mime_type=self.mime_type,
            binary=self.is_binary,
            suffix=self.suffix,
            base64_string=base64_string,
        )

        if self.filename is not None:
            fs.write(self.filename)
            return None
        return fs

    @abstractmethod
    def format_data(self) -> Union[str, bytes]:
        """Convert the input data to the target format."""
        pass

    def export(self) -> Optional["FileStore"]:
        """Export the data to a FileStore instance."""
        formatted_data = self.format_data()
        return self._create_filestore(formatted_data)


class JSONLExport(FileExport):
    mime_type = "application/jsonl"
    suffix = "jsonl"
    is_binary = False

    def format_data(self) -> str:
        output = io.StringIO()
        for entry in self.data:
            key, values = list(entry.items())[0]
            output.write(f'{{"{key}": {values}}}\n')
        return output.getvalue()


class TabularExport(FileExport, ABC):
    """Base class for exports that use tabular data."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.header, self.rows = self.data._get_tabular_data(
            remove_prefix=self.remove_prefix, pretty_labels=self.pretty_labels
        )


class CSVExport(TabularExport):
    mime_type = "text/csv"
    suffix = "csv"
    is_binary = False

    def format_data(self) -> str:
        output = io.StringIO()
        writer = csv.writer(output)
        writer.writerow(self.header)
        writer.writerows(self.rows)
        return output.getvalue()


class ExcelExport(TabularExport):
    mime_type = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
    suffix = "xlsx"
    is_binary = True

    def __init__(self, *args, sheet_name: Optional[str] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.sheet_name = sheet_name or "Results"

    def format_data(self) -> bytes:
        wb = Workbook()
        ws = wb.active
        ws.title = self.sheet_name

        # Write header
        for col, value in enumerate(self.header, 1):
            ws.cell(row=1, column=col, value=value)

        # Write data rows
        for row_idx, row_data in enumerate(self.rows, 2):
            for col, value in enumerate(row_data, 1):
                ws.cell(row=row_idx, column=col, value=value)

        # Save to bytes buffer
        buffer = io.BytesIO()
        wb.save(buffer)
        buffer.seek(0)
        return buffer.getvalue()


import sqlite3
from typing import Any


class SQLiteExport(TabularExport):
    mime_type = "application/x-sqlite3"
    suffix = "db"
    is_binary = True

    def __init__(
        self, *args, table_name: str = "results", if_exists: str = "replace", **kwargs
    ):
        """
        Initialize SQLite export.

        Args:
            table_name: Name of the table to create
            if_exists: How to handle existing table ('fail', 'replace', or 'append')
        """
        super().__init__(*args, **kwargs)
        self.table_name = table_name
        self.if_exists = if_exists

    def _get_column_types(self) -> list[tuple[str, str]]:
        """Infer SQL column types from the data."""
        column_types = []

        # Check first row of data for types
        if self.rows:
            first_row = self.rows[0]
            for header, value in zip(self.header, first_row):
                if isinstance(value, bool):
                    sql_type = "BOOLEAN"
                elif isinstance(value, int):
                    sql_type = "INTEGER"
                elif isinstance(value, float):
                    sql_type = "REAL"
                else:
                    sql_type = "TEXT"
                column_types.append((header, sql_type))
        else:
            # If no data, default to TEXT
            column_types = [(header, "TEXT") for header in self.header]

        return column_types

    def _create_table(self, cursor: sqlite3.Cursor) -> None:
        """Create the table with appropriate schema."""
        column_types = self._get_column_types()

        # Drop existing table if replace mode
        if self.if_exists == "replace":
            cursor.execute(f"DROP TABLE IF EXISTS {self.table_name}")
        elif self.if_exists == "fail":
            cursor.execute(
                f"SELECT name FROM sqlite_master WHERE type='table' AND name=?",
                (self.table_name,),
            )
            if cursor.fetchone():
                raise ValueError(f"Table {self.table_name} already exists")

        # Create table
        columns = ", ".join(f'"{col}" {dtype}' for col, dtype in column_types)
        create_table_sql = f"""
        CREATE TABLE IF NOT EXISTS {self.table_name} (
            {columns}
        )
        """
        cursor.execute(create_table_sql)

    def format_data(self) -> bytes:
        """Convert the data to a SQLite database file."""
        buffer = io.BytesIO()

        # Create in-memory database
        conn = sqlite3.connect(":memory:")
        cursor = conn.cursor()

        # Create table and insert data
        self._create_table(cursor)

        # Prepare placeholders for INSERT
        placeholders = ",".join(["?" for _ in self.header])
        insert_sql = f"INSERT INTO {self.table_name} ({','.join(self.header)}) VALUES ({placeholders})"

        # Insert data
        cursor.executemany(insert_sql, self.rows)
        conn.commit()

        # Save to file buffer
        conn.backup(sqlite3.connect(buffer))
        conn.close()

        buffer.seek(0)
        return buffer.getvalue()

    def _validate_params(self) -> None:
        """Validate initialization parameters."""
        valid_if_exists = {"fail", "replace", "append"}
        if self.if_exists not in valid_if_exists:
            raise ValueError(
                f"if_exists must be one of {valid_if_exists}, got {self.if_exists}"
            )

        # Validate table name (basic SQLite identifier validation)
        if not self.table_name.isalnum() and not all(c in "_" for c in self.table_name):
            raise ValueError(
                f"Invalid table name: {self.table_name}. Must contain only alphanumeric characters and underscores."
            )



import os
from edsl.exceptions.general import MissingAPIKeyError


class JobsChecks:
    def __init__(self, jobs):
        """Checks a Jobs object for missing API keys and other requirements."""
        self.jobs = jobs

    def check_api_keys(self) -> None:
        from edsl.language_models.model import Model

        if len(self.jobs.models) == 0:
            models = [Model()]
        else:
            models = self.jobs.models

        for model in models:  # + [Model()]:
            if not model.has_valid_api_key():
                raise MissingAPIKeyError(
                    model_name=str(model.model),
                    inference_service=model._inference_service_,
                )

    def get_missing_api_keys(self) -> set:
        """
        Returns a list of the api keys that a user needs to run this job, but does not currently have in their .env file.
        """
        missing_api_keys = set()

        from edsl.language_models.model import Model
        from edsl.enums import service_to_api_keyname

        for model in self.jobs.models + [Model()]:
            if not model.has_valid_api_key():
                key_name = service_to_api_keyname.get(
                    model._inference_service_, "NOT FOUND"
                )
                missing_api_keys.add(key_name)

        return missing_api_keys

    def user_has_ep_api_key(self) -> bool:
        """
        Returns True if the user has an EXPECTED_PARROT_API_KEY in their env.

        Otherwise, returns False.
        """

        coop_api_key = os.getenv("EXPECTED_PARROT_API_KEY")

        if coop_api_key is not None:
            return True
        else:
            return False

    def user_has_all_model_keys(self):
        """
        Returns True if the user has all model keys required to run their job.

        Otherwise, returns False.
        """

        try:
            self.check_api_keys()
            return True
        except MissingAPIKeyError:
            return False
        except Exception:
            raise

    def needs_external_llms(self) -> bool:
        """
        Returns True if the job needs external LLMs to run.

        Otherwise, returns False.
        """
        # These cases are necessary to skip the API key check during doctests

        # Accounts for Results.example()
        all_agents_answer_questions_directly = len(self.jobs.agents) > 0 and all(
            [hasattr(a, "answer_question_directly") for a in self.jobs.agents]
        )

        # Accounts for InterviewExceptionEntry.example()
        only_model_is_test = set([m.model for m in self.jobs.models]) == set(["test"])

        # Accounts for Survey.__call__
        all_questions_are_functional = set(
            [q.question_type for q in self.jobs.survey.questions]
        ) == set(["functional"])

        if (
            all_agents_answer_questions_directly
            or only_model_is_test
            or all_questions_are_functional
        ):
            return False
        else:
            return True

    def needs_key_process(self):
        """
        A User needs the key process when:
        1. They don't have all the model keys
        2. They don't have the EP API
        3. They need external LLMs to run the job
        """
        return (
            not self.user_has_all_model_keys()
            and not self.user_has_ep_api_key()
            and self.needs_external_llms()
        )

    def status(self) -> dict:
        """
        Returns a dictionary with the status of the job checks.
        """
        return {
            "user_has_ep_api_key": self.user_has_ep_api_key(),
            "user_has_all_model_keys": self.user_has_all_model_keys(),
            "needs_external_llms": self.needs_external_llms(),
            "needs_key_process": self.needs_key_process(),
        }

    def key_process(self):
        import secrets
        from dotenv import load_dotenv
        from edsl.config import CONFIG
        from edsl.coop.coop import Coop
        from edsl.utilities.utilities import write_api_key_to_env

        missing_api_keys = self.get_missing_api_keys()

        edsl_auth_token = secrets.token_urlsafe(16)

        print("You're missing some of the API keys needed to run this job:")
        for api_key in missing_api_keys:
            print(f"     🔑 {api_key}")
        print(
            "\nYou can either add the missing keys to your .env file, or use remote inference."
        )
        print("Remote inference allows you to run jobs on our server.")

        coop = Coop()
        coop._display_login_url(
            edsl_auth_token=edsl_auth_token,
            link_description="\n🚀 To use remote inference, sign up at the following link:",
        )

        print(
            "\nOnce you log in, we will automatically retrieve your Expected Parrot API key and continue your job remotely."
        )

        api_key = coop._poll_for_api_key(edsl_auth_token)

        if api_key is None:
            print("\nTimed out waiting for login. Please try again.")
            return

        path_to_env = write_api_key_to_env(api_key)
        print("\n✨ API key retrieved and written to .env file at the following path:")
        print(f"    {path_to_env}")

        # Retrieve API key so we can continue running the job
        load_dotenv()


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from typing import Optional, TYPE_CHECKING, Protocol
import sys
from edsl.scenarios.FileStore import HTMLFileStore
from edsl.config import CONFIG
from edsl.coop.coop import Coop


class ResultsProtocol(Protocol):
    """Protocol defining the required interface for Results objects."""

    @property
    def has_unfixed_exceptions(self) -> bool: ...

    @property
    def task_history(self) -> "TaskHistoryProtocol": ...


class TaskHistoryProtocol(Protocol):
    """Protocol defining the required interface for TaskHistory objects."""

    @property
    def indices(self) -> list: ...

    def html(self, cta: str, open_in_browser: bool, return_link: bool) -> str: ...


class RunParametersProtocol(Protocol):
    """Protocol defining the required interface for RunParameters objects."""

    @property
    def print_exceptions(self) -> bool: ...


class ResultsExceptionsHandler:
    """Handles exception reporting and display functionality."""

    def __init__(
        self, results: ResultsProtocol, parameters: RunParametersProtocol
    ) -> None:
        self.results = results
        self.parameters = parameters

        self.open_in_browser = self._get_browser_setting()
        self.remote_logging = self._get_remote_logging_setting()

    def _get_browser_setting(self) -> bool:
        """Determine if exceptions should be opened in browser based on config."""
        setting = CONFIG.get("EDSL_OPEN_EXCEPTION_REPORT_URL")
        if setting == "True":
            return True
        elif setting == "False":
            return False
        else:
            raise Exception(
                "EDSL_OPEN_EXCEPTION_REPORT_URL must be either True or False"
            )

    def _get_remote_logging_setting(self) -> bool:
        """Get remote logging setting from coop."""
        try:
            coop = Coop()
            return coop.edsl_settings["remote_logging"]
        except Exception as e:
            # print(e)
            return False

    def _generate_error_message(self, indices) -> str:
        """Generate appropriate error message based on number of exceptions."""
        msg = f"Exceptions were raised.\n" 
        return msg

    def handle_exceptions(self) -> None:
        """Handle exceptions by printing messages and generating reports as needed."""
        if not (
            self.results.has_unfixed_exceptions and self.parameters.print_exceptions
        ):
            return

        # Print error message
        error_msg = self._generate_error_message(self.results.task_history.indices)
        print(error_msg, file=sys.stderr)

        # Generate HTML report
        filepath = self.results.task_history.html(
            open_in_browser=self.open_in_browser,
            return_link=True,
        )

        # Handle remote logging if enabled
        if self.remote_logging:
            filestore = HTMLFileStore(filepath)
            coop_details = filestore.push(description="Exceptions Report")
            print(coop_details)



from typing import Literal, Optional, Type, Union

from edsl.agents.Agent import Agent
from edsl.agents.AgentList import AgentList
from edsl.data.Cache import Cache
from edsl.language_models.ModelList import ModelList
from edsl.notebooks.Notebook import Notebook
from edsl.results.Results import Results
from edsl.scenarios.Scenario import Scenario
from edsl.scenarios.ScenarioList import ScenarioList
from edsl.surveys.Survey import Survey
from edsl.study.Study import Study

from edsl.language_models.LanguageModel import LanguageModel
from edsl.questions.QuestionBase import QuestionBase

EDSLObject = Union[
    Agent,
    AgentList,
    Cache,
    LanguageModel,
    ModelList,
    Notebook,
    Type[QuestionBase],
    Results,
    Scenario,
    ScenarioList,
    Survey,
    Study,
]

ObjectType = Literal[
    "agent",
    "agent_list",
    "cache",
    "model",
    "model_list",
    "notebook",
    "question",
    "results",
    "scenario",
    "scenario_list",
    "survey",
    "study",
]


RemoteJobStatus = Literal[
    "queued",
    "running",
    "completed",
    "failed",
]

VisibilityType = Literal[
    "private",
    "public",
    "unlisted",
]


class ObjectRegistry:
    """
    Utility class to map object types to database models.
    """

    objects = [
        {"object_type": "agent", "edsl_class": Agent},
        {"object_type": "agent_list", "edsl_class": AgentList},
        {"object_type": "cache", "edsl_class": Cache},
        {"object_type": "model", "edsl_class": LanguageModel},
        {"object_type": "model_list", "edsl_class": ModelList},
        {"object_type": "notebook", "edsl_class": Notebook},
        {"object_type": "question", "edsl_class": QuestionBase},
        {"object_type": "results", "edsl_class": Results},
        {"object_type": "scenario", "edsl_class": Scenario},
        {"object_type": "scenario_list", "edsl_class": ScenarioList},
        {"object_type": "survey", "edsl_class": Survey},
        {"object_type": "study", "edsl_class": Study},
    ]
    object_type_to_edsl_class = {o["object_type"]: o["edsl_class"] for o in objects}
    edsl_class_to_object_type = {
        o["edsl_class"].__name__: o["object_type"] for o in objects
    }

    @classmethod
    def get_object_type_by_edsl_class(cls, edsl_object: EDSLObject) -> ObjectType:
        if isinstance(edsl_object, type):
            edsl_class_name = edsl_object.__name__
        else:
            edsl_class_name = type(edsl_object).__name__
        if edsl_class_name.startswith("Question"):
            edsl_class_name = "QuestionBase"
        object_type = cls.edsl_class_to_object_type.get(edsl_class_name)
        if object_type is None:
            raise ValueError(f"Object type not found for {edsl_object=}")
        return object_type

    @classmethod
    def get_edsl_class_by_object_type(cls, object_type: ObjectType) -> EDSLObject:
        EDSL_object = cls.object_type_to_edsl_class.get(object_type)
        if EDSL_object is None:
            raise ValueError(f"EDSL class not found for {object_type=}")
        return EDSL_object

    @classmethod
    def get_registry(
        cls,
        subclass_registry: Optional[dict] = None,
        exclude_classes: Optional[list] = None,
    ) -> dict:
        """
        Return the registry of objects.

        Exclude objects that are already registered in subclass_registry.
        This allows the user to isolate Coop-only objects.

        Also exclude objects if their class name is in the exclude_classes list.
        """

        if subclass_registry is None:
            subclass_registry = {}
        if exclude_classes is None:
            exclude_classes = []

        return {
            class_name: o["edsl_class"]
            for o in cls.objects
            if (class_name := o["edsl_class"].__name__) not in subclass_registry
            and class_name not in exclude_classes
        }



from typing import Union, List, Any, Optional
import asyncio
import time
from threading import RLock
from edsl.jobs.decorators import synchronized_class

from typing import Union, List, Any, Optional
import asyncio
import time
from threading import RLock
from edsl.jobs.decorators import synchronized_class


@synchronized_class
class TokenBucket:
    """This is a token bucket used to respect rate limits to services.
    It can operate either locally or remotely via a REST API based on initialization parameters.
    """

    def __new__(
        cls,
        *,
        bucket_name: str,
        bucket_type: str,
        capacity: Union[int, float],
        refill_rate: Union[int, float],
        remote_url: Optional[str] = None,
    ):
        """Factory method to create either a local or remote token bucket.

        Args:
            bucket_name: Name of the bucket
            bucket_type: Type of the bucket
            capacity: Maximum number of tokens
            refill_rate: Rate at which tokens are refilled
            remote_url: If provided, creates a remote token bucket client
        """
        if remote_url is not None:
            # Import here to avoid circular imports
            from edsl.jobs.buckets.TokenBucketClient import TokenBucketClient

            return TokenBucketClient(
                bucket_name=bucket_name,
                bucket_type=bucket_type,
                capacity=capacity,
                refill_rate=refill_rate,
                api_base_url=remote_url,
            )

        # Create a local token bucket
        instance = super(TokenBucket, cls).__new__(cls)
        return instance

    def __init__(
        self,
        *,
        bucket_name: str,
        bucket_type: str,
        capacity: Union[int, float],
        refill_rate: Union[int, float],
        remote_url: Optional[str] = None,
    ):
        # Skip initialization if this is a remote bucket
        if remote_url is not None:
            return

        self.bucket_name = bucket_name
        self.bucket_type = bucket_type
        self.capacity = capacity
        self.added_tokens = 0
        self._lock = RLock()

        self.target_rate = (
            capacity * 60
        )  # set this here because it can change with turbo mode

        self._old_capacity = capacity
        self.tokens = capacity  # Current number of available tokens
        self.refill_rate = refill_rate  # Rate at which tokens are refilled
        self._old_refill_rate = refill_rate
        self.last_refill = time.monotonic()  # Last refill time
        self.log: List[Any] = []
        self.turbo_mode = False

        self.creation_time = time.monotonic()

        self.num_requests = 0
        self.num_released = 0
        self.tokens_returned = 0

    def turbo_mode_on(self):
        """Set the refill rate to infinity."""
        if self.turbo_mode:
            pass
        else:
            # pass
            self.turbo_mode = True
            self.capacity = float("inf")
            self.refill_rate = float("inf")

    def turbo_mode_off(self):
        """Restore the refill rate to its original value."""
        self.turbo_mode = False
        self.capacity = self._old_capacity
        self.refill_rate = self._old_refill_rate

    def __add__(self, other) -> "TokenBucket":
        """Combine two token buckets.

        The resulting bucket has the minimum capacity and refill rate of the two buckets.
        This is useful, for example, if we have two calls to the same model on the same service but have different temperatures.
        """
        return TokenBucket(
            bucket_name=self.bucket_name,
            bucket_type=self.bucket_type,
            capacity=min(self.capacity, other.capacity),
            refill_rate=min(self.refill_rate, other.refill_rate),
        )

    def __repr__(self):
        return f"TokenBucket(bucket_name={self.bucket_name}, bucket_type='{self.bucket_type}', capacity={self.capacity}, refill_rate={self.refill_rate})"

    def add_tokens(self, tokens: Union[int, float]) -> None:
        """Add tokens to the bucket, up to the maximum capacity.

        :param tokens: The number of tokens to add to the bucket.

        >>> bucket = TokenBucket(bucket_name="test", bucket_type="test", capacity=10, refill_rate=1)
        >>> bucket.tokens
        10
        >>> bucket.add_tokens(5)
        >>> bucket.tokens
        10
        """
        self.tokens_returned += tokens
        self.tokens = min(self.capacity, self.tokens + tokens)
        self.log.append((time.monotonic(), self.tokens))

    def refill(self) -> None:
        """Refill the bucket with new tokens based on elapsed time.



        >>> bucket = TokenBucket(bucket_name="test", bucket_type="test", capacity=10, refill_rate=1)
        >>> bucket.tokens = 0
        >>> bucket.refill()
        >>> bucket.tokens > 0
        True
        """
        """Refill the bucket with new tokens based on elapsed time."""
        now = time.monotonic()
        # print(f"Time is now: {now}; Last refill time: {self.last_refill}")
        elapsed = now - self.last_refill
        # print("Elapsed time: ", elapsed)
        refill_amount = elapsed * self.refill_rate
        self.tokens = min(self.capacity, self.tokens + refill_amount)
        self.last_refill = now

        if self.tokens < self.capacity:
            pass
            # print(f"Refilled. Current tokens: {self.tokens:.4f}")
            # print(f"Elapsed time: {elapsed:.4f} seconds")
            # print(f"Refill amount: {refill_amount:.4f}")

        self.log.append((now, self.tokens))

    def wait_time(self, requested_tokens: Union[float, int]) -> float:
        """Calculate the time to wait for the requested number of tokens."""
        # self.refill()  # Update the current token count
        if self.tokens >= requested_tokens:
            return 0
        return (requested_tokens - self.tokens) / self.refill_rate

    async def get_tokens(
        self, amount: Union[int, float] = 1, cheat_bucket_capacity=True
    ) -> None:
        """Wait for the specified number of tokens to become available.


        :param amount: The number of tokens
        :param warn: If True, warn if the requested amount exceeds the bucket capacity.

        >>> bucket = TokenBucket(bucket_name="test", bucket_type="test", capacity=10, refill_rate=1)
        >>> asyncio.run(bucket.get_tokens(5))
        >>> bucket.tokens
        5
        >>> asyncio.run(bucket.get_tokens(9))
        >>> bucket.tokens < 1
        True

        >>> bucket = TokenBucket(bucket_name="test", bucket_type="test", capacity=10, refill_rate=1)
        >>> asyncio.run(bucket.get_tokens(11, cheat_bucket_capacity=False))
        Traceback (most recent call last):
        ...
        ValueError: Requested amount exceeds bucket capacity. Bucket capacity: 10, requested amount: 11. As the bucket never overflows, the requested amount will never be available.
        >>> asyncio.run(bucket.get_tokens(11, cheat_bucket_capacity=True))
        >>> bucket.capacity
        12.100000000000001
        """
        self.num_requests += amount
        if amount >= self.capacity:
            if not cheat_bucket_capacity:
                msg = f"Requested amount exceeds bucket capacity. Bucket capacity: {self.capacity}, requested amount: {amount}. As the bucket never overflows, the requested amount will never be available."
                raise ValueError(msg)
            else:
                self.capacity = amount * 1.10
                self._old_capacity = self.capacity

        start_time = time.monotonic()
        while True:
            self.refill()  # Refill based on elapsed time
            if self.tokens >= amount:
                self.tokens -= amount
                break

            wait_time = self.wait_time(amount)
            if wait_time > 0:
                await asyncio.sleep(wait_time)

        self.num_released += amount
        now = time.monotonic()
        self.log.append((now, self.tokens))
        return None

    def get_log(self) -> list[tuple]:
        return self.log

    def visualize(self):
        """Visualize the token bucket over time."""
        times, tokens = zip(*self.get_log())
        start_time = times[0]
        times = [t - start_time for t in times]  # Normalize time to start from 0
        from matplotlib import pyplot as plt

        plt.figure(figsize=(10, 6))
        plt.plot(times, tokens, label="Tokens Available")
        plt.xlabel("Time (seconds)", fontsize=12)
        plt.ylabel("Number of Tokens", fontsize=12)
        details = f"{self.bucket_name} ({self.bucket_type}) Bucket Usage Over Time\nCapacity: {self.capacity:.1f}, Refill Rate: {self.refill_rate:.1f}/second"
        plt.title(details, fontsize=14)

        plt.legend()
        plt.grid(True)
        plt.tight_layout()
        plt.show()

    def get_throughput(self, time_window: Optional[float] = None) -> float:
        """
        Calculate the empirical bucket throughput in tokens per minute for the specified time window.

        :param time_window: The time window in seconds to calculate the throughput for.
        :return: The throughput in tokens per minute.

        >>> bucket = TokenBucket(bucket_name="test", bucket_type="test", capacity=100, refill_rate=10)
        >>> asyncio.run(bucket.get_tokens(50))
        >>> time.sleep(1)  # Wait for 1 second
        >>> asyncio.run(bucket.get_tokens(30))
        >>> throughput = bucket.get_throughput(1)
        >>> 4750 < throughput < 4850
        True
        """
        now = time.monotonic()

        if time_window is None:
            start_time = self.creation_time
        else:
            start_time = now - time_window

        if start_time < self.creation_time:
            start_time = self.creation_time

        elapsed_time = now - start_time

        if elapsed_time == 0:
            return self.num_released / 0.001

        return (self.num_released / elapsed_time) * 60


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from textwrap import dedent
from typing import Optional


class LanguageModelExceptions(Exception):
    explanation = (
        "This is the base class for all exceptions in the LanguageModel class."
    )

    def __init__(self, message):
        super().__init__(message)
        self.message = message


class LanguageModelNoResponseError(LanguageModelExceptions):
    explanation = (
        """This happens when the LLM API cannot be reached and/or does not respond."""
    )

    def __init__(self, message):
        super().__init__(message)


class LanguageModelBadResponseError(LanguageModelExceptions):
    explanation = """This happens when the LLM API can be reached and responds, does not return a usable answer."""

    def __init__(self, message, response_json: Optional[dict] = None):
        super().__init__(message)
        self.response_json = response_json


class LanguageModelNotFound(LanguageModelExceptions):
    def __init__(self, model_name):
        msg = dedent(
            f"""\
            Model {model_name} not found.
            To create an instance of this model, pass the model name to a `Model` object.
            You can optionally pass additional parameters to the model, e.g.: 
            >>> m = Model('gpt-4-1106-preview', temperature=0.5)
            
            To use the default model, simply run your job without specifying a model.
            To check the default model, run the following code:
            >>> Model()

            To see information about all available models, run the following code:
            >>> Model.available()

            See https://docs.expectedparrot.com/en/latest/language_models.html#available-models for more details.
            """
        )
        super().__init__(msg)


class LanguageModelResponseNotJSONError(LanguageModelExceptions):
    pass


class LanguageModelMissingAttributeError(LanguageModelExceptions):
    pass


class LanguageModelAttributeTypeError(LanguageModelExceptions):
    pass


class LanguageModelDoNotAddError(LanguageModelExceptions):
    pass



from typing import Union, Optional, List, Generator, Dict
from edsl.utilities.remove_edsl_version import remove_edsl_version
from edsl.Base import RepresentationMixin


class Instruction(RepresentationMixin):
    def __init__(
        self, name, text, preamble="You were given the following instructions:"
    ):
        self.name = name
        self.text = text
        self.preamble = preamble

    def __str__(self):
        return self.text

    def __repr__(self):
        return """Instruction(name="{}", text="{}")""".format(self.name, self.text)

    @classmethod
    def example(cls) -> "Instruction":
        return cls(name="example", text="This is an example instruction.")

    def to_dict(self, add_edsl_version=True):
        d = {
            "name": self.name,
            "text": self.text,
            "edsl_class_name": "Instruction",
            "preamble": self.preamble,
        }
        if add_edsl_version:
            from edsl import __version__

            d["edsl_version"] = __version__
            d["edsl_class_name"] = "Instruction"
        return d

    def add_question(self, question) -> "Survey":
        from edsl.surveys.Survey import Survey

        return Survey([self, question])

    def __hash__(self) -> int:
        """Return a hash of the question."""
        from edsl.utilities.utilities import dict_hash

        return dict_hash(self.to_dict(add_edsl_version=False))

    @classmethod
    @remove_edsl_version
    def from_dict(cls, data):
        return cls(
            data["name"],
            data["text"],
            data.get("preamble", "You were given the following instructions:"),
        )



import aiohttp
import json
import requests

from typing import Any, Optional, Union, Literal, TypedDict
from uuid import UUID
from collections import UserDict, defaultdict

import edsl
from pathlib import Path

from edsl.config import CONFIG
from edsl.data.CacheEntry import CacheEntry
from edsl.jobs.Jobs import Jobs
from edsl.surveys.Survey import Survey

from edsl.exceptions.coop import (
    CoopInvalidURLError,
    CoopNoUUIDError,
    CoopServerResponseError,
)
from edsl.coop.utils import (
    EDSLObject,
    ObjectRegistry,
    ObjectType,
    RemoteJobStatus,
    VisibilityType,
)

from edsl.coop.CoopFunctionsMixin import CoopFunctionsMixin
from edsl.coop.ExpectedParrotKeyHandler import ExpectedParrotKeyHandler

from edsl.inference_services.data_structures import ServiceToModelsMapping


class RemoteInferenceResponse(TypedDict):
    job_uuid: str
    results_uuid: str
    results_url: str
    latest_error_report_uuid: str
    latest_error_report_url: str
    status: str
    reason: str
    credits_consumed: float
    version: str


class RemoteInferenceCreationInfo(TypedDict):
    uuid: str
    description: str
    status: str
    iterations: int
    visibility: str
    version: str


class Coop(CoopFunctionsMixin):
    """
    Client for the Expected Parrot API.
    """

    def __init__(
        self, api_key: Optional[str] = None, url: Optional[str] = None
    ) -> None:
        """
        Initialize the client.
        - Provide an API key directly, or through an env variable.
        - Provide a URL directly, or use the default one.
        """
        self.ep_key_handler = ExpectedParrotKeyHandler()
        self.api_key = api_key or self.ep_key_handler.get_ep_api_key()

        self.url = url or CONFIG.EXPECTED_PARROT_URL
        if self.url.endswith("/"):
            self.url = self.url[:-1]
        if "chick.expectedparrot" in self.url:
            self.api_url = "https://chickapi.expectedparrot.com"
        elif "expectedparrot" in self.url:
            self.api_url = "https://api.expectedparrot.com"
        elif "localhost:1234" in self.url:
            self.api_url = "http://localhost:8000"
        else:
            self.api_url = self.url
        self._edsl_version = edsl.__version__

    def get_progress_bar_url(self):
        return f"{CONFIG.EXPECTED_PARROT_URL}"

    ################
    # BASIC METHODS
    ################
    @property
    def headers(self) -> dict:
        """
        Return the headers for the request.
        """
        headers = {}
        if self.api_key:
            headers["Authorization"] = f"Bearer {self.api_key}"
        else:
            headers["Authorization"] = f"Bearer None"
        return headers

    def _send_server_request(
        self,
        uri: str,
        method: str,
        payload: Optional[dict[str, Any]] = None,
        params: Optional[dict[str, Any]] = None,
        timeout: Optional[float] = 5,
    ) -> requests.Response:
        """
        Send a request to the server and return the response.
        """
        url = f"{self.api_url}/{uri}"
        method = method.upper()
        if payload is None:
            timeout = 40
        elif (
            method.upper() == "POST"
            and "json_string" in payload
            and payload.get("json_string") is not None
        ):
            timeout = max(40, (len(payload.get("json_string", "")) // (1024 * 1024)))
        try:
            if method in ["GET", "DELETE"]:
                response = requests.request(
                    method, url, params=params, headers=self.headers, timeout=timeout
                )
            elif method in ["POST", "PATCH"]:
                response = requests.request(
                    method,
                    url,
                    params=params,
                    json=payload,
                    headers=self.headers,
                    timeout=timeout,
                )
            else:
                raise Exception(f"Invalid {method=}.")
        except requests.ConnectionError:
            raise requests.ConnectionError(f"Could not connect to the server at {url}.")

        return response

    def _get_latest_stable_version(self, version: str) -> str:
        """
        Extract the latest stable PyPI version from a version string.

        Examples:
        - Decrement the patch number of a dev version: "0.1.38.dev1" -> "0.1.37"
        - Return a stable version as is: "0.1.37" -> "0.1.37"
        """
        if "dev" not in version:
            return version
        else:
            # For 0.1.38.dev1, split into ["0", "1", "38", "dev1"]
            major, minor, patch = version.split(".")[:3]

            current_patch = int(patch)
            latest_patch = current_patch - 1
            return f"{major}.{minor}.{latest_patch}"

    def _user_version_is_outdated(
        self, user_version_str: str, server_version_str: str
    ) -> bool:
        """
        Check if the user's EDSL version is outdated compared to the server's.
        """
        server_stable_version_str = self._get_latest_stable_version(server_version_str)
        user_stable_version_str = self._get_latest_stable_version(user_version_str)

        # Turn the version strings into tuples of ints for comparison
        user_stable_version = tuple(map(int, user_stable_version_str.split(".")))
        server_stable_version = tuple(map(int, server_stable_version_str.split(".")))

        return user_stable_version < server_stable_version

    def _resolve_server_response(
        self, response: requests.Response, check_api_key: bool = True
    ) -> None:
        """
        Check the response from the server and raise errors as appropriate.
        """
        # Get EDSL version from header
        # breakpoint()
        server_edsl_version = response.headers.get("X-EDSL-Version")

        if server_edsl_version:
            if self._user_version_is_outdated(
                user_version_str=self._edsl_version,
                server_version_str=server_edsl_version,
            ):
                print(
                    "Please upgrade your EDSL version to access our latest features. To upgrade, open your terminal and run `pip install --upgrade edsl`"
                )

        if response.status_code >= 400:
            try:
                message = response.json().get("detail")
            except json.JSONDecodeError:
                raise CoopServerResponseError(
                    f"Server returned status code {response.status_code}."
                    "JSON response could not be decoded.",
                    "The server response was: " + response.text,
                )
            # print(response.text)
            if "The API key you provided is invalid" in message and check_api_key:
                import secrets
                from edsl.utilities.utilities import write_api_key_to_env

                edsl_auth_token = secrets.token_urlsafe(16)

                print("Your Expected Parrot API key is invalid.")
                self._display_login_url(
                    edsl_auth_token=edsl_auth_token,
                    link_description="\n🔗 Use the link below to log in to Expected Parrot so we can automatically update your API key.",
                )
                api_key = self._poll_for_api_key(edsl_auth_token)

                if api_key is None:
                    print("\nTimed out waiting for login. Please try again.")
                    return

                print("\n✨ API key retrieved.")

                if stored_in_user_space := self.ep_key_handler.ask_to_store(api_key):
                    pass
                else:
                    path_to_env = write_api_key_to_env(api_key)
                    print(
                        "\n✨ API key retrieved and written to .env file at the following path:"
                    )
                    print(f"    {path_to_env}")
                    print("Rerun your code to try again with a valid API key.")
                return

            elif "Authorization" in message:
                print(message)
                message = "Please provide an Expected Parrot API key."

            raise CoopServerResponseError(message)

    def _poll_for_api_key(
        self, edsl_auth_token: str, timeout: int = 120
    ) -> Union[str, None]:
        """
        Allows the user to retrieve their Expected Parrot API key by logging in with an EDSL auth token.

        :param edsl_auth_token: The EDSL auth token to use for login
        :param timeout: Maximum time to wait for login, in seconds (default: 120)
        """
        import time
        from datetime import datetime

        start_poll_time = time.time()
        waiting_for_login = True
        while waiting_for_login:
            elapsed_time = time.time() - start_poll_time
            if elapsed_time > timeout:
                # Timed out waiting for the user to log in
                print("\r" + " " * 80 + "\r", end="")
                return None

            api_key = self._get_api_key(edsl_auth_token)
            if api_key is not None:
                print("\r" + " " * 80 + "\r", end="")
                return api_key
            else:
                duration = 5
                time_checked = datetime.now().strftime("%Y-%m-%d %I:%M:%S %p")
                frames = ["⠋", "⠙", "⠹", "⠸", "⠼", "⠴", "⠦", "⠧", "⠇", "⠏"]
                start_time = time.time()
                i = 0
                while time.time() - start_time < duration:
                    print(
                        f"\r{frames[i % len(frames)]} Waiting for login. Last checked: {time_checked}",
                        end="",
                        flush=True,
                    )
                    time.sleep(0.1)
                    i += 1

    def _json_handle_none(self, value: Any) -> Any:
        """
        Handle None values during JSON serialization.
        - Return "null" if the value is None. Otherwise, don't return anything.
        """
        if value is None:
            return "null"

    def _resolve_uuid_or_alias(
        self, uuid: Union[str, UUID] = None, url: str = None
    ) -> tuple[Optional[str], Optional[str], Optional[str]]:
        """
        Resolve the uuid or alias information from a uuid or a url.
        Returns a tuple of (uuid, owner_username, alias)
        - For content/<uuid> URLs: returns (uuid, None, None)
        - For content/<username>/<alias> URLs: returns (None, username, alias)
        """
        if not url and not uuid:
            raise CoopNoUUIDError("No uuid or url provided for the object.")

        if not uuid and url:
            parts = (
                url.replace("http://", "")
                .replace("https://", "")
                .rstrip("/")
                .split("/")
            )

            # Remove domain
            parts = parts[1:]

            if len(parts) < 2 or parts[0] != "content":
                raise CoopInvalidURLError(
                    f"Invalid URL format. The URL must end with /content/<uuid> or /content/<username>/<alias>: {url}"
                )

            if len(parts) == 2:
                obj_uuid = parts[1]
                return obj_uuid, None, None
            elif len(parts) == 3:
                username, alias = parts[1], parts[2]
                return None, username, alias
            else:
                raise CoopInvalidURLError(
                    f"Invalid URL format. The URL must end with /content/<uuid> or /content/<username>/<alias>: {url}"
                )

        return str(uuid), None, None

    @property
    def edsl_settings(self) -> dict:
        """
        Retrieve and return the EDSL settings stored on Coop.
        If no response is received within 5 seconds, return an empty dict.
        """
        from requests.exceptions import Timeout

        try:
            response = self._send_server_request(
                uri="api/v0/edsl-settings", method="GET", timeout=5
            )
            self._resolve_server_response(response, check_api_key=False)
            return response.json()
        except Timeout:
            return {}

    ################
    # Objects
    ################
    def create(
        self,
        object: EDSLObject,
        description: Optional[str] = None,
        alias: Optional[str] = None,
        visibility: Optional[VisibilityType] = "unlisted",
    ) -> dict:
        """
        Create an EDSL object in the Coop server.
        """
        object_type = ObjectRegistry.get_object_type_by_edsl_class(object)
        response = self._send_server_request(
            uri=f"api/v0/object",
            method="POST",
            payload={
                "description": description,
                "alias": alias,
                "json_string": json.dumps(
                    object.to_dict(),
                    default=self._json_handle_none,
                ),
                "object_type": object_type,
                "visibility": visibility,
                "version": self._edsl_version,
            },
        )
        self._resolve_server_response(response)
        response_json = response.json()
        return {
            "description": response_json.get("description"),
            "object_type": object_type,
            "url": f"{self.url}/content/{response_json.get('uuid')}",
            "uuid": response_json.get("uuid"),
            "version": self._edsl_version,
            "visibility": response_json.get("visibility"),
        }

    def get(
        self,
        uuid: Union[str, UUID] = None,
        url: str = None,
        expected_object_type: Optional[ObjectType] = None,
    ) -> EDSLObject:
        """
        Retrieve an EDSL object by its uuid/url or by owner username and alias.
        - If the object's visibility is private, the user must be the owner.
        - Optionally, check if the retrieved object is of a certain type.

        :param uuid: the uuid of the object either in str or UUID format.
        :param url: the url of the object (can be content/uuid or content/username/alias format).
        :param expected_object_type: the expected type of the object.

        :return: the object instance.
        """
        obj_uuid, owner_username, alias = self._resolve_uuid_or_alias(uuid, url)

        if obj_uuid:
            response = self._send_server_request(
                uri=f"api/v0/object",
                method="GET",
                params={"uuid": obj_uuid},
            )
        else:
            response = self._send_server_request(
                uri=f"api/v0/object/alias",
                method="GET",
                params={"owner_username": owner_username, "alias": alias},
            )

        self._resolve_server_response(response)
        json_string = response.json().get("json_string")
        object_type = response.json().get("object_type")
        if expected_object_type and object_type != expected_object_type:
            raise Exception(f"Expected {expected_object_type=} but got {object_type=}")
        edsl_class = ObjectRegistry.object_type_to_edsl_class.get(object_type)
        object = edsl_class.from_dict(json.loads(json_string))
        return object

    def get_all(self, object_type: ObjectType) -> list[dict[str, Any]]:
        """
        Retrieve all objects of a certain type associated with the user.
        """
        edsl_class = ObjectRegistry.object_type_to_edsl_class.get(object_type)
        response = self._send_server_request(
            uri=f"api/v0/objects",
            method="GET",
            params={"type": object_type},
        )
        self._resolve_server_response(response)
        objects = [
            {
                "object": edsl_class.from_dict(json.loads(o.get("json_string"))),
                "uuid": o.get("uuid"),
                "version": o.get("version"),
                "description": o.get("description"),
                "visibility": o.get("visibility"),
                "url": f"{self.url}/content/{o.get('uuid')}",
            }
            for o in response.json()
        ]
        return objects

    def delete(self, uuid: Union[str, UUID] = None, url: str = None) -> dict:
        """
        Delete an object from the server.
        """
        obj_uuid, _, _ = self._resolve_uuid_or_alias(uuid, url)
        response = self._send_server_request(
            uri=f"api/v0/object",
            method="DELETE",
            params={"uuid": obj_uuid},
        )

        self._resolve_server_response(response)
        return response.json()

    def patch(
        self,
        uuid: Union[str, UUID] = None,
        url: str = None,
        description: Optional[str] = None,
        alias: Optional[str] = None,
        value: Optional[EDSLObject] = None,
        visibility: Optional[VisibilityType] = None,
    ) -> dict:
        """
        Change the attributes of an uploaded object
        - Only supports visibility for now
        """
        if description is None and visibility is None and value is None:
            raise Exception("Nothing to patch.")
        obj_uuid, _, _ = self._resolve_uuid_or_alias(uuid, url)
        response = self._send_server_request(
            uri=f"api/v0/object",
            method="PATCH",
            params={"uuid": obj_uuid},
            payload={
                "description": description,
                "alias": alias,
                "json_string": (
                    json.dumps(
                        value.to_dict(),
                        default=self._json_handle_none,
                    )
                    if value
                    else None
                ),
                "visibility": visibility,
            },
        )
        self._resolve_server_response(response)
        return response.json()

    ################
    # Remote Cache
    ################
    def remote_cache_create(
        self,
        cache_entry: CacheEntry,
        visibility: VisibilityType = "private",
        description: Optional[str] = None,
    ) -> dict:
        """
        Create a single remote cache entry.
        If an entry with the same key already exists in the database, update it instead.

        :param cache_entry: The cache entry to send to the server.
        :param visibility: The visibility of the cache entry.
        :param optional description: A description for this entry in the remote cache.

        >>> entry = CacheEntry.example()
        >>> coop.remote_cache_create(cache_entry=entry)
        {'status': 'success', 'created_entry_count': 1, 'updated_entry_count': 0}
        """
        response = self._send_server_request(
            uri="api/v0/remote-cache",
            method="POST",
            payload={
                "json_string": json.dumps(cache_entry.to_dict()),
                "version": self._edsl_version,
                "visibility": visibility,
                "description": description,
            },
        )
        self._resolve_server_response(response)
        response_json = response.json()
        created_entry_count = response_json.get("created_entry_count", 0)
        if created_entry_count > 0:
            self.remote_cache_create_log(
                response,
                description="Upload new cache entries to server",
                cache_entry_count=created_entry_count,
            )
        return response.json()

    def remote_cache_create_many(
        self,
        cache_entries: list[CacheEntry],
        visibility: VisibilityType = "private",
        description: Optional[str] = None,
    ) -> dict:
        """
        Create many remote cache entries.
        If an entry with the same key already exists in the database, update it instead.

        :param cache_entries: The list of cache entries to send to the server.
        :param visibility: The visibility of the cache entries.
        :param optional description: A description for these entries in the remote cache.

        >>> entries = [CacheEntry.example(randomize=True) for _ in range(10)]
        >>> coop.remote_cache_create_many(cache_entries=entries)
        {'status': 'success', 'created_entry_count': 10, 'updated_entry_count': 0}
        """
        payload = [
            {
                "json_string": json.dumps(c.to_dict()),
                "version": self._edsl_version,
                "visibility": visibility,
                "description": description,
            }
            for c in cache_entries
        ]
        response = self._send_server_request(
            uri="api/v0/remote-cache/many",
            method="POST",
            payload=payload,
            timeout=40,
        )
        self._resolve_server_response(response)
        response_json = response.json()
        created_entry_count = response_json.get("created_entry_count", 0)
        if created_entry_count > 0:
            self.remote_cache_create_log(
                response,
                description="Upload new cache entries to server",
                cache_entry_count=created_entry_count,
            )
        return response.json()

    def remote_cache_get(
        self,
        exclude_keys: Optional[list[str]] = None,
        select_keys: Optional[list[str]] = None,
    ) -> list[CacheEntry]:
        """
        Get all remote cache entries.

        :param optional exclude_keys: Exclude CacheEntry objects with these keys.

        >>> coop.remote_cache_get()
        [CacheEntry(...), CacheEntry(...), ...]
        """
        if exclude_keys is None:
            exclude_keys = []
        if select_keys is None:
            select_keys = []
        response = self._send_server_request(
            uri="api/v0/remote-cache/get-many",
            method="POST",
            payload={"keys": exclude_keys, "selected_keys": select_keys},
            timeout=40,
        )
        self._resolve_server_response(response)
        return [
            CacheEntry.from_dict(json.loads(v.get("json_string")))
            for v in response.json()
        ]

    def remote_cache_get_diff(
        self,
        client_cacheentry_keys: list[str],
    ) -> dict:
        """
        Get the difference between local and remote cache entries for a user.
        """
        response = self._send_server_request(
            uri="api/v0/remote-cache/get-diff",
            method="POST",
            payload={"keys": client_cacheentry_keys},
            timeout=40,
        )
        self._resolve_server_response(response)
        response_json = response.json()
        response_dict = {
            "client_missing_cacheentries": [
                CacheEntry.from_dict(json.loads(c.get("json_string")))
                for c in response_json.get("client_missing_cacheentries", [])
            ],
            "server_missing_cacheentry_keys": response_json.get(
                "server_missing_cacheentry_keys", []
            ),
        }
        downloaded_entry_count = len(response_dict["client_missing_cacheentries"])
        if downloaded_entry_count > 0:
            self.remote_cache_create_log(
                response,
                description="Download missing cache entries to client",
                cache_entry_count=downloaded_entry_count,
            )
        return response_dict

    def remote_cache_clear(self) -> dict:
        """
        Clear all remote cache entries.

        >>> entries = [CacheEntry.example(randomize=True) for _ in range(10)]
        >>> coop.remote_cache_create_many(cache_entries=entries)
        >>> coop.remote_cache_clear()
        {'status': 'success', 'deleted_entry_count': 10}
        """
        response = self._send_server_request(
            uri="api/v0/remote-cache/delete-all",
            method="DELETE",
        )
        self._resolve_server_response(response)
        response_json = response.json()
        deleted_entry_count = response_json.get("deleted_entry_count", 0)
        if deleted_entry_count > 0:
            self.remote_cache_create_log(
                response,
                description="Clear cache entries",
                cache_entry_count=deleted_entry_count,
            )
        return response.json()

    def remote_cache_create_log(
        self, response: requests.Response, description: str, cache_entry_count: int
    ) -> Union[dict, None]:
        """
        If a remote cache action has been completed successfully,
        log the action.
        """
        if 200 <= response.status_code < 300:
            log_response = self._send_server_request(
                uri="api/v0/remote-cache-log",
                method="POST",
                payload={
                    "description": description,
                    "cache_entry_count": cache_entry_count,
                },
            )
            self._resolve_server_response(log_response)
            return response.json()

    def remote_cache_clear_log(self) -> dict:
        """
        Clear all remote cache log entries.

        >>> coop.remote_cache_clear_log()
        {'status': 'success'}
        """
        response = self._send_server_request(
            uri="api/v0/remote-cache-log/delete-all",
            method="DELETE",
        )
        self._resolve_server_response(response)
        return response.json()

    def remote_inference_create(
        self,
        job: Jobs,
        description: Optional[str] = None,
        status: RemoteJobStatus = "queued",
        visibility: Optional[VisibilityType] = "unlisted",
        initial_results_visibility: Optional[VisibilityType] = "unlisted",
        iterations: Optional[int] = 1,
    ) -> RemoteInferenceCreationInfo:
        """
        Send a remote inference job to the server.

        :param job: The EDSL job to send to the server.
        :param optional description: A description for this entry in the remote cache.
        :param status: The status of the job. Should be 'queued', unless you are debugging.
        :param visibility: The visibility of the cache entry.
        :param iterations: The number of times to run each interview.

        >>> job = Jobs.example()
        >>> coop.remote_inference_create(job=job, description="My job")
        {'uuid': '9f8484ee-b407-40e4-9652-4133a7236c9c', 'description': 'My job', 'status': 'queued', 'iterations': None, 'visibility': 'unlisted', 'version': '0.1.38.dev1'}
        """
        response = self._send_server_request(
            uri="api/v0/remote-inference",
            method="POST",
            payload={
                "json_string": json.dumps(
                    job.to_dict(),
                    default=self._json_handle_none,
                ),
                "description": description,
                "status": status,
                "iterations": iterations,
                "visibility": visibility,
                "version": self._edsl_version,
                "initial_results_visibility": initial_results_visibility,
            },
        )
        self._resolve_server_response(response)
        response_json = response.json()

        return RemoteInferenceCreationInfo(
            **{
                "uuid": response_json.get("job_uuid"),
                "description": response_json.get("description"),
                "status": response_json.get("status"),
                "iterations": response_json.get("iterations"),
                "visibility": response_json.get("visibility"),
                "version": self._edsl_version,
            }
        )

    def remote_inference_get(
        self, job_uuid: Optional[str] = None, results_uuid: Optional[str] = None
    ) -> RemoteInferenceResponse:
        """
        Get the details of a remote inference job.
        You can pass either the job uuid or the results uuid as a parameter.
        If you pass both, the job uuid will be prioritized.

        :param job_uuid: The UUID of the EDSL job.
        :param results_uuid: The UUID of the results associated with the EDSL job.

        >>> coop.remote_inference_get("9f8484ee-b407-40e4-9652-4133a7236c9c")
        {'job_uuid': '9f8484ee-b407-40e4-9652-4133a7236c9c', 'results_uuid': 'dd708234-31bf-4fe1-8747-6e232625e026', 'results_url': 'https://www.expectedparrot.com/content/dd708234-31bf-4fe1-8747-6e232625e026', 'latest_error_report_uuid': None, 'latest_error_report_url': None, 'status': 'completed', 'reason': None, 'credits_consumed': 0.35, 'version': '0.1.38.dev1'}
        """
        if job_uuid is None and results_uuid is None:
            raise ValueError("Either job_uuid or results_uuid must be provided.")
        elif job_uuid is not None:
            params = {"job_uuid": job_uuid}
        else:
            params = {"results_uuid": results_uuid}

        response = self._send_server_request(
            uri="api/v0/remote-inference",
            method="GET",
            params=params,
        )
        self._resolve_server_response(response)
        data = response.json()

        results_uuid = data.get("results_uuid")
        latest_error_report_uuid = data.get("latest_error_report_uuid")

        if results_uuid is None:
            results_url = None
        else:
            results_url = f"{self.url}/content/{results_uuid}"

        if latest_error_report_uuid is None:
            latest_error_report_url = None
        else:
            latest_error_report_url = (
                f"{self.url}/home/remote-inference/error/{latest_error_report_uuid}"
            )

        return RemoteInferenceResponse(
            **{
                "job_uuid": data.get("job_uuid"),
                "results_uuid": results_uuid,
                "results_url": results_url,
                "latest_error_report_uuid": latest_error_report_uuid,
                "latest_error_report_url": latest_error_report_url,
                "status": data.get("status"),
                "reason": data.get("reason"),
                "credits_consumed": data.get("price"),
                "version": data.get("version"),
            }
        )

    def get_running_jobs(self) -> list[str]:
        """
        Get a list of currently running job IDs.

        Returns:
            list[str]: List of running job UUIDs
        """
        response = self._send_server_request(uri="jobs/status", method="GET")
        self._resolve_server_response(response)
        return response.json().get("running_jobs", [])

    def remote_inference_cost(
        self, input: Union[Jobs, Survey], iterations: int = 1
    ) -> int:
        """
        Get the cost of a remote inference job.

        :param input: The EDSL job to send to the server.

        >>> job = Jobs.example()
        >>> coop.remote_inference_cost(input=job)
        {'credits': 0.77, 'usd': 0.0076950000000000005}
        """
        if isinstance(input, Jobs):
            job = input
        elif isinstance(input, Survey):
            job = Jobs(survey=input)
        else:
            raise TypeError("Input must be either a Job or a Survey.")

        response = self._send_server_request(
            uri="api/v0/remote-inference/cost",
            method="POST",
            payload={
                "json_string": json.dumps(
                    job.to_dict(),
                    default=self._json_handle_none,
                ),
                "iterations": iterations,
            },
        )
        self._resolve_server_response(response)
        response_json = response.json()
        return {
            "credits": response_json.get("cost_in_credits"),
            "usd": response_json.get("cost_in_usd"),
        }

    ################
    # DUNDER METHODS
    ################
    def __repr__(self):
        """Return a string representation of the client."""
        return f"Client(api_key='{self.api_key}', url='{self.url}')"

    ################
    # EXPERIMENTAL
    ################
    async def remote_async_execute_model_call(
        self, model_dict: dict, user_prompt: str, system_prompt: str
    ) -> dict:
        url = self.api_url + "/inference/"
        # print("Now using url: ", url)
        data = {
            "model_dict": model_dict,
            "user_prompt": user_prompt,
            "system_prompt": system_prompt,
        }
        # Use aiohttp to send a POST request asynchronously
        async with aiohttp.ClientSession() as session:
            async with session.post(url, json=data) as response:
                response_data = await response.json()
        return response_data

    def web(
        self,
        survey: dict,
        platform: Literal[
            "google_forms", "lime_survey", "survey_monkey"
        ] = "lime_survey",
        email=None,
    ):
        url = f"{self.api_url}/api/v0/export_to_{platform}"
        if email:
            data = {"json_string": json.dumps({"survey": survey, "email": email})}
        else:
            data = {"json_string": json.dumps({"survey": survey, "email": ""})}

        response_json = requests.post(url, headers=self.headers, data=json.dumps(data))

        return response_json

    def fetch_prices(self) -> dict:
        """
        Fetch model prices from Coop. If the request fails, return an empty dict.
        """

        from edsl.coop.PriceFetcher import PriceFetcher

        from edsl.config import CONFIG

        if CONFIG.get("EDSL_FETCH_TOKEN_PRICES") == "True":
            price_fetcher = PriceFetcher()
            return price_fetcher.fetch_prices()
        elif CONFIG.get("EDSL_FETCH_TOKEN_PRICES") == "False":
            return {}
        else:
            raise ValueError(
                "Invalid EDSL_FETCH_TOKEN_PRICES value---should be 'True' or 'False'."
            )

    def fetch_models(self) -> ServiceToModelsMapping:
        """
        Fetch a dict of available models from Coop.

        Each key in the dict is an inference service, and each value is a list of models from that service.
        """
        response = self._send_server_request(uri="api/v0/models", method="GET")
        self._resolve_server_response(response)
        data = response.json()
        return ServiceToModelsMapping(data)

    def fetch_working_models(self) -> list[dict]:
        """
        Fetch a list of working models from Coop.

        Example output:

        [
            {
                "service": "openai",
                "model": "gpt-4o",
                "works_with_text": True,
                "works_with_images": True,
                "usd_per_1M_input_tokens": 2.5,
                "usd_per_1M_output_tokens": 10.0,
            }
        ]
        """
        response = self._send_server_request(uri="api/v0/working-models", method="GET")
        self._resolve_server_response(response)
        data = response.json()
        return [
            {
                "service": record.get("service"),
                "model": record.get("model"),
                "works_with_text": record.get("works_with_text"),
                "works_with_images": record.get("works_with_images"),
                "usd_per_1M_input_tokens": record.get("input_price_per_1M_tokens"),
                "usd_per_1M_output_tokens": record.get("output_price_per_1M_tokens"),
            }
            for record in data
        ]

    def fetch_rate_limit_config_vars(self) -> dict:
        """
        Fetch a dict of rate limit config vars from Coop.

        The dict keys are RPM and TPM variables like EDSL_SERVICE_RPM_OPENAI.
        """
        response = self._send_server_request(
            uri="api/v0/config-vars",
            method="GET",
        )
        self._resolve_server_response(response)
        data = response.json()
        return data

    def _display_login_url(
        self, edsl_auth_token: str, link_description: Optional[str] = None
    ):
        """
        Uses rich.print to display a login URL.

        - We need this function because URL detection with print() does not work alongside animations in VSCode.
        """
        from rich import print as rich_print

        url = f"{CONFIG.EXPECTED_PARROT_URL}/login?edsl_auth_token={edsl_auth_token}"

        if link_description:
            rich_print(
                f"{link_description}\n    [#38bdf8][link={url}]{url}[/link][/#38bdf8]"
            )
        else:
            rich_print(f"    [#38bdf8][link={url}]{url}[/link][/#38bdf8]")

    def _get_api_key(self, edsl_auth_token: str):
        """
        Given an EDSL auth token, find the corresponding user's API key.
        """

        response = self._send_server_request(
            uri="api/v0/get-api-key",
            method="POST",
            payload={
                "edsl_auth_token": edsl_auth_token,
            },
        )
        data = response.json()
        api_key = data.get("api_key")
        return api_key

    def login(self):
        """
        Starts the EDSL auth token login flow.
        """
        import secrets
        from dotenv import load_dotenv
        from edsl.utilities.utilities import write_api_key_to_env

        edsl_auth_token = secrets.token_urlsafe(16)

        self._display_login_url(
            edsl_auth_token=edsl_auth_token,
            link_description="\n🔗 Use the link below to log in to Expected Parrot so we can automatically update your API key.",
        )
        api_key = self._poll_for_api_key(edsl_auth_token)

        if api_key is None:
            raise Exception("Timed out waiting for login. Please try again.")

        path_to_env = write_api_key_to_env(api_key)
        print("\n✨ API key retrieved and written to .env file at the following path:")
        print(f"    {path_to_env}")

        # Add API key to environment
        load_dotenv()


def main():
    """
    A simple example for the coop client
    """
    from uuid import uuid4
    from edsl import (
        Agent,
        AgentList,
        Cache,
        Notebook,
        QuestionFreeText,
        QuestionMultipleChoice,
        Results,
        Scenario,
        ScenarioList,
        Survey,
    )
    from edsl.coop import Coop
    from edsl.data.CacheEntry import CacheEntry
    from edsl.jobs import Jobs

    # init & basics
    API_KEY = "b"
    coop = Coop(api_key=API_KEY)
    coop
    coop.edsl_settings

    ##############
    # A. A simple example
    ##############
    # .. create and manipulate an object through the Coop client
    response = coop.create(QuestionMultipleChoice.example())
    coop.get(uuid=response.get("uuid"))
    coop.get(uuid=response.get("uuid"), expected_object_type="question")
    coop.get(url=response.get("url"))
    coop.create(QuestionMultipleChoice.example())
    coop.get_all("question")
    coop.patch(uuid=response.get("uuid"), visibility="private")
    coop.patch(uuid=response.get("uuid"), description="hey")
    coop.patch(uuid=response.get("uuid"), value=QuestionFreeText.example())
    # coop.patch(uuid=response.get("uuid"), value=Survey.example()) - should throw error
    coop.get(uuid=response.get("uuid"))
    coop.delete(uuid=response.get("uuid"))

    # .. create and manipulate an object through the class
    response = QuestionMultipleChoice.example().push()
    QuestionMultipleChoice.pull(uuid=response.get("uuid"))
    QuestionMultipleChoice.pull(url=response.get("url"))
    QuestionMultipleChoice.patch(uuid=response.get("uuid"), visibility="private")
    QuestionMultipleChoice.patch(uuid=response.get("uuid"), description="hey")
    QuestionMultipleChoice.patch(
        uuid=response.get("uuid"), value=QuestionFreeText.example()
    )
    QuestionMultipleChoice.pull(response.get("uuid"))
    QuestionMultipleChoice.delete(response.get("uuid"))

    ##############
    # B. Examples with all objects
    ##############
    OBJECTS = [
        ("agent", Agent),
        ("agent_list", AgentList),
        ("cache", Cache),
        ("notebook", Notebook),
        ("question", QuestionMultipleChoice),
        ("results", Results),
        ("scenario", Scenario),
        ("scenario_list", ScenarioList),
        ("survey", Survey),
    ]
    for object_type, cls in OBJECTS:
        print(f"Testing {object_type} objects")
        # 1. Delete existing objects
        existing_objects = coop.get_all(object_type)
        for item in existing_objects:
            coop.delete(uuid=item.get("uuid"))
        # 2. Create new objects
        example = cls.example()
        response_1 = coop.create(example)
        response_2 = coop.create(cls.example(), visibility="private")
        response_3 = coop.create(cls.example(), visibility="public")
        response_4 = coop.create(
            cls.example(), visibility="unlisted", description="hey"
        )
        # 3. Retrieve all objects
        objects = coop.get_all(object_type)
        assert len(objects) == 4
        # 4. Try to retrieve an item that does not exist
        try:
            coop.get(uuid=uuid4())
        except Exception as e:
            print(e)
        # 5. Try to retrieve all test objects by their uuids
        for response in [response_1, response_2, response_3, response_4]:
            coop.get(uuid=response.get("uuid"))
        # 6. Change visibility of all objects
        for item in objects:
            coop.patch(uuid=item.get("uuid"), visibility="private")
        # 6. Change description of all objects
        for item in objects:
            coop.patch(uuid=item.get("uuid"), description="hey")
        # 7. Delete all objects
        for item in objects:
            coop.delete(uuid=item.get("uuid"))
        assert len(coop.get_all(object_type)) == 0

    ##############
    # C. Remote Cache
    ##############
    # clear
    coop.remote_cache_clear()
    assert coop.remote_cache_get() == []
    # create one remote cache entry
    cache_entry = CacheEntry.example()
    cache_entry.to_dict()
    coop.remote_cache_create(cache_entry)
    # create many remote cache entries
    cache_entries = [CacheEntry.example(randomize=True) for _ in range(10)]
    coop.remote_cache_create_many(cache_entries)
    # get all remote cache entries
    coop.remote_cache_get()
    coop.remote_cache_get(exclude_keys=[])
    coop.remote_cache_get(exclude_keys=["a"])
    exclude_keys = [cache_entry.key for cache_entry in cache_entries]
    coop.remote_cache_get(exclude_keys)
    # clear
    coop.remote_cache_clear()
    coop.remote_cache_get()

    ##############
    # D. Remote Inference
    ##############
    job = Jobs.example()
    coop.remote_inference_cost(job)
    job_coop_object = coop.remote_inference_create(job)
    job_coop_results = coop.remote_inference_get(job_coop_object.get("uuid"))
    coop.get(uuid=job_coop_results.get("results_uuid"))



from typing import Optional


class SmartInt(int):
    pass


class SmartFloat(float):
    pass


class SmartStr(str):
    def clipboard(self) -> None:
        try:
            import pyperclip
        except ImportError:
            print(
                "pyperclip is not installed. Run `pip install pyperclip` to install it."
            )
            return None

        pyperclip.copy(self)
        print("Text copied to clipboard.")

    def write(self, filename: str):
        with open(filename, "w") as f:
            f.write(str(self))
        return None

    def _repr_html_(self):
        pass

    def markdown(self):
        return SmartMarkdown(self)

    def pdf(self, filename: Optional[str] = None):  # Markdown will have this as well
        # renders the markdown as a pdf that can be downloaded
        from edsl.results.MarkdownToPDF import MarkdownToPDF

        return MarkdownToPDF(self, filename).preview()

    def docx(self, filename: Optional[str] = None):
        # renders the markdown as a docx that can be downloaded
        from edsl.results.MarkdownToDocx import MarkdownToDocx

        return MarkdownToDocx(self, filename).preview()

    def edit(self):
        from edsl.results.TextEditor import TextEditor

        editor = TextEditor(self)
        self = self.__class__(editor.edit_gui())
        # print(f"Updated text: {self}")


class SmartMarkdown(SmartStr):
    def _repr_markdown_(self):
        return self

    def _repr_html_(self):
        from IPython.display import Markdown, display

        display(Markdown(self))


class SmartLaTeX(SmartStr):
    def _repr_html_(self):
        print(self)

    def pdf(self, filename: Optional[str] = None):
        from edsl.results.LaTeXToPDF import LaTeXToPDF

        return LaTeXToPDF(self, filename).preview()

    def docx(self, filename: Optional[str] = None):
        from edsl.results.LaTeXToDocx import LaTeXToDocx

        return LaTeXToDocx(self, filename).preview()

    def edit(self):
        from edsl.results.TextEditor import TextEditor

        editor = TextEditor(self)
        self = self.__class__(editor.edit_gui())
        # print(f"Updated LaTeX: {self}")


class FirstObject:
    def __new__(self, value):
        if isinstance(value, int):
            return SmartInt(value)
        if isinstance(value, float):
            return SmartFloat(value)
        if isinstance(value, str):
            return SmartStr(value)
        return value



from typing import Dict, List, Set
from warnings import warn
import logging
from edsl.prompts.Prompt import Prompt

from edsl.agents.QuestionTemplateReplacementsBuilder import (
    QuestionTemplateReplacementsBuilder as QTRB,
)


class QuestionInstructionPromptBuilder:
    """Handles the construction and rendering of question instructions."""

    def __init__(self, prompt_constructor: "PromptConstructor"):
        self.prompt_constructor = prompt_constructor

        self.model = self.prompt_constructor.model
        self.survey = self.prompt_constructor.survey
        self.question = self.prompt_constructor.question

    def build(self) -> Prompt:
        """Builds the complete question instructions prompt with all necessary components.

        Returns:
            Prompt: The fully rendered question instructions
        """
        import time
        
        start = time.time()
        
        # Create base prompt
        base_start = time.time()
        base_prompt = self._create_base_prompt()
        base_end = time.time()
        logging.debug(f"Time for base prompt: {base_end - base_start}")
        
        # Enrich with options
        enrich_start = time.time()
        enriched_prompt = self._enrich_with_question_options(base_prompt)
        enrich_end = time.time()
        logging.debug(f"Time for enriching with options: {enrich_end - enrich_start}")
        
        # Render prompt
        render_start = time.time()
        rendered_prompt = self._render_prompt(enriched_prompt)
        render_end = time.time()
        logging.debug(f"Time for rendering prompt: {render_end - render_start}")
        
        # Validate template variables
        validate_start = time.time()
        self._validate_template_variables(rendered_prompt)
        validate_end = time.time()
        logging.debug(f"Time for template validation: {validate_end - validate_start}")
        
        # Append survey instructions
        append_start = time.time()
        final_prompt = self._append_survey_instructions(rendered_prompt)
        append_end = time.time()
        logging.debug(f"Time for appending survey instructions: {append_end - append_start}")
        
        end = time.time()
        logging.debug(f"Total time in build_question_instructions: {end - start}")
        
        return final_prompt

    def _create_base_prompt(self) -> Dict:
        """Creates the initial prompt with basic question data.

        Returns:
            Dict: Base question data
        """
        return {
            "prompt": Prompt(self.question.get_instructions(model=self.model.model)),
            "data": self.question.data.copy(),
        }

    def _enrich_with_question_options(self, prompt_data: Dict) -> Dict:
        """Enriches the prompt data with question options if they exist.

        Args:
            prompt_data: Dictionary containing prompt and question data

        Returns:
            Dict: Enriched prompt data
        """
        import time
        
        start = time.time()
        
        if "question_options" in prompt_data["data"]:
            from edsl.agents.question_option_processor import QuestionOptionProcessor
            
            processor_start = time.time()
            question_options = QuestionOptionProcessor(
                self.prompt_constructor
            ).get_question_options(question_data=prompt_data["data"])
            processor_end = time.time()
            logging.debug(f"Time to process question options: {processor_end - processor_start}")
            
            prompt_data["data"]["question_options"] = question_options
            
        end = time.time()
        logging.debug(f"Total time in _enrich_with_question_options: {end - start}")
        
        return prompt_data

    def _render_prompt(self, prompt_data: Dict) -> Prompt:
        """Renders the prompt using the replacement dictionary.

        Args:
            prompt_data: Dictionary containing prompt and question data

        Returns:
            Prompt: Rendered instructions
        """
        import time
        
        start = time.time()
        
        # Build replacement dict
        dict_start = time.time()
        replacement_dict = QTRB(self.prompt_constructor).build_replacement_dict(
            prompt_data["data"]
        )
        dict_end = time.time()
        logging.debug(f"Time to build replacement dict: {dict_end - dict_start}")
        
        # Render with dict
        render_start = time.time()
        result = prompt_data["prompt"].render(replacement_dict)
        render_end = time.time()
        logging.debug(f"Time to render with dict: {render_end - render_start}")
        
        end = time.time()
        logging.debug(f"Total time in _render_prompt: {end - start}")
        
        return result

    def _validate_template_variables(self, rendered_prompt: Prompt) -> None:
        """Validates that all template variables have been properly replaced.

        Args:
            rendered_prompt: The rendered prompt to validate

        Warns:
            If any template variables remain undefined
        """
        undefined_vars = rendered_prompt.undefined_template_variables({})

        # Check for question names in undefined variables
        self._check_question_names_in_undefined_vars(undefined_vars)

        # Warn about any remaining undefined variables
        if undefined_vars:
            warn(f"Question instructions still has variables: {undefined_vars}.")

    def _check_question_names_in_undefined_vars(self, undefined_vars: Set[str]) -> None:
        """Checks if any undefined variables match question names in the survey.

        Args:
            undefined_vars: Set of undefined template variables
        """
        for question_name in self.survey.question_names:
            if question_name in undefined_vars:
                logging.warning(f"Question name found in undefined_template_variables: {question_name}")

    def _append_survey_instructions(self, rendered_prompt: Prompt) -> Prompt:
        """Appends any relevant survey instructions to the rendered prompt.

        Args:
            rendered_prompt: The rendered prompt to append instructions to

        Returns:
            Prompt: Final prompt with survey instructions
        """
        relevant_instructions = self.survey._relevant_instructions(
            self.question.question_name
        )

        if not relevant_instructions:
            return rendered_prompt

        preamble = Prompt(text="")
        for instruction in relevant_instructions:
            preamble += instruction.text

        return preamble + rendered_prompt



import warnings
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from edsl.surveys.Survey import Survey
    from edsl.scenarios.ScenarioList import ScenarioList


class CheckSurveyScenarioCompatibility:

    def __init__(self, survey: "Survey", scenarios: "ScenarioList"):
        self.survey = survey
        self.scenarios = scenarios

    def check(self, strict: bool = False, warn: bool = False) -> None:
        """Check if the parameters in the survey and scenarios are consistent.

        >>> from edsl.jobs.Jobs import Jobs
        >>> from edsl.questions.QuestionFreeText import QuestionFreeText
        >>> from edsl.surveys.Survey import Survey
        >>> from edsl.scenarios.Scenario import Scenario
        >>> q = QuestionFreeText(question_text = "{{poo}}", question_name = "ugly_question")
        >>> j = Jobs(survey = Survey(questions=[q]))
        >>> cs = CheckSurveyScenarioCompatibility(j.survey, j.scenarios)
        >>> with warnings.catch_warnings(record=True) as w:
        ...     cs.check(warn = True)
        ...     assert len(w) == 1
        ...     assert issubclass(w[-1].category, UserWarning)
        ...     assert "The following parameters are in the survey but not in the scenarios" in str(w[-1].message)

        >>> q = QuestionFreeText(question_text = "{{poo}}", question_name = "ugly_question")
        >>> s = Scenario({'plop': "A", 'poo': "B"})
        >>> j = Jobs(survey = Survey(questions=[q])).by(s)
        >>> cs = CheckSurveyScenarioCompatibility(j.survey, j.scenarios)
        >>> cs.check(strict = True)
        Traceback (most recent call last):
        ...
        ValueError: The following parameters are in the scenarios but not in the survey: {'plop'}

        >>> q = QuestionFreeText(question_text = "Hello", question_name = "ugly_question")
        >>> s = Scenario({'ugly_question': "B"})
        >>> from edsl.scenarios.ScenarioList import ScenarioList
        >>> cs = CheckSurveyScenarioCompatibility(Survey(questions=[q]), ScenarioList([s]))
        >>> cs.check()
        Traceback (most recent call last):
        ...
        ValueError: The following names are in both the survey question_names and the scenario keys: {'ugly_question'}. This will create issues.
        """
        survey_parameters: set = self.survey.parameters
        scenario_parameters: set = self.scenarios.parameters

        msg0, msg1, msg2 = None, None, None

        # look for key issues
        if intersection := set(self.scenarios.parameters) & set(
            self.survey.question_names
        ):
            msg0 = f"The following names are in both the survey question_names and the scenario keys: {intersection}. This will create issues."

            raise ValueError(msg0)

        if in_survey_but_not_in_scenarios := survey_parameters - scenario_parameters:
            msg1 = f"The following parameters are in the survey but not in the scenarios: {in_survey_but_not_in_scenarios}"
        if in_scenarios_but_not_in_survey := scenario_parameters - survey_parameters:
            msg2 = f"The following parameters are in the scenarios but not in the survey: {in_scenarios_but_not_in_survey}"

        if msg1 or msg2:
            message = "\n".join(filter(None, [msg1, msg2]))
            if strict:
                raise ValueError(message)
            else:
                if warn:
                    warnings.warn(message)

        if self.scenarios.has_jinja_braces:
            warnings.warn(
                "The scenarios have Jinja braces ({{ and }}). Converting to '<<' and '>>'. If you want a different conversion, use the convert_jinja_braces method first to modify the scenario."
            )
            self.scenarios = self.scenarios._convert_jinja_braces()


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from textwrap import dedent
from dataclasses import dataclass

from typing import List, Dict

from edsl.auto.StageBase import StageBase
from edsl.auto.StageBase import FlowDataBase

from edsl.auto.StagePersonaDimensions import StagePersonaDimensions
from edsl import Model
from edsl.questions import QuestionList, QuestionExtract
from edsl.scenarios import Scenario

from edsl.auto.utilities import gen_pipeline


class StagePersonaDimensionValues(StageBase):
    input = StagePersonaDimensions.output

    @dataclass
    class Output(FlowDataBase):
        attribute_results: List[str]
        dimension_values: Dict[str, str]
        persona: str

    output = Output

    def handle_data(self, data):
        attribute_results = data.attribute_results
        persona = data.persona
        m = Model()
        q = QuestionExtract(
            question_text=dedent(
                """\
            This is a persona: "{{ persona }}"
            They vary on the following dimensions: "{{ attribute_results }}"
            For each dimenion, what are some values that this persona might have for that dimension?
            Please keep answers very short, ideally one word.
            """
            ),
            answer_template={k: None for k in attribute_results},
            question_name="dimension_values",
        )
        results = (
            q.by(Scenario({"attribute_results": attribute_results, "persona": persona}))
            .by(m)
            .run()
        )
        results.select("attribute_results", "dimension_values").print()
        return self.output(
            dimension_values=results.select("dimension_values").first(),
            attribute_results=attribute_results,
            persona=persona,
        )


if __name__ == "__main__":
    from edsl.auto.StageQuestions import StageQuestions
    from edsl.auto.StagePersona import StagePersona
    from edsl.auto.StagePersonaDimensions import StagePersonaDimensions

    pipeline = gen_pipeline(
        [
            StageQuestions,
            StagePersona,
            StagePersonaDimensions,
            StagePersonaDimensionValues,
        ]
    )
    pipeline.process(
        pipeline.input(
            overall_question="What are some factors that could determine whether someone likes ice cream?"
        )
    )



from typing import (
    Protocol,
    List,
    Any,
    Optional,
    TYPE_CHECKING,
    Sequence,
    Union,
    Literal,
)

if TYPE_CHECKING:
    from edsl.results.Dataset import Dataset

from edsl.results.table_data_class import TableData

from edsl.results.table_renderers import DataTablesRenderer, PandasStyleRenderer

Row = Sequence[Union[str, int, float, bool, None]]
TableFormat = Literal[
    "grid", "simple", "pipe", "orgtbl", "rst", "mediawiki", "html", "latex"
]


class TableRenderer(Protocol):
    """Table renderer protocol"""

    def render_html(self, table_data: TableData) -> str:
        pass


# Modified TableDisplay class
class TableDisplay:
    def __init__(
        self,
        headers: Sequence[str],
        data: Sequence[Row],
        tablefmt: Optional[TableFormat] = None,
        raw_data_set: "Dataset" = None,
        renderer_class: Optional[TableRenderer] = None,
    ):
        assert len(headers) == len(data[0])  # Check if headers and data are consistent

        self.headers = headers
        self.data = data
        self.tablefmt = tablefmt
        self.raw_data_set = raw_data_set

        self.renderer_class = renderer_class or PandasStyleRenderer

        # Handle printing parameters from raw_data_set
        if hasattr(raw_data_set, "print_parameters"):
            self.printing_parameters = (
                raw_data_set.print_parameters if raw_data_set.print_parameters else {}
            )
        else:
            self.printing_parameters = {}

    def _repr_html_(self) -> str:
        table_data = TableData(
            headers=self.headers,
            data=self.data,
            parameters=self.printing_parameters,
            raw_data_set=self.raw_data_set,
        )
        return self.renderer_class(table_data).render_html()

    def __repr__(self):
        from tabulate import tabulate

        return tabulate(self.data, headers=self.headers, tablefmt=self.tablefmt)

    @classmethod
    def from_dictionary(
        cls,
        dictionary: dict,
        tablefmt: Optional[TableFormat] = None,
        renderer: Optional[TableRenderer] = None,
    ) -> "TableDisplay":
        headers = list(dictionary.keys())
        data = [list(dictionary.values())]
        return cls(headers, data, tablefmt, renderer_class=renderer)

    @classmethod
    def from_dictionary_wide(
        cls,
        dictionary: dict,
        tablefmt: Optional[TableFormat] = None,
        renderer: Optional[TableRenderer] = None,
    ) -> "TableDisplay":
        headers = ["key", "value"]
        data = [[k, v] for k, v in dictionary.items()]
        return cls(headers, data, tablefmt, renderer_class=renderer)

    @classmethod
    def from_dataset(
        cls,
        dataset: "Dataset",
        tablefmt: Optional[TableFormat] = None,
        renderer: Optional[TableRenderer] = None,
    ) -> "TableDisplay":
        headers, data = dataset._tabular()
        return cls(headers, data, tablefmt, dataset, renderer_class=renderer)

    def long(self) -> "TableDisplay":
        """Convert to long format"""
        new_header = ["row", "key", "value"]
        new_data = []
        for index, row in enumerate(self.data):
            new_data.extend([[index, k, v] for k, v in zip(self.headers, row)])
        return TableDisplay(
            new_header, new_data, self.tablefmt, renderer_class=self.renderer_class
        )


# Example usage:
if __name__ == "__main__":
    headers = ["Name", "Age", "City"]
    data = [["John", 30, "New York"], ["Jane", 25, "London"]]

    # Using default (Pandas) renderer
    table1 = TableDisplay(headers, data)

    # Using DataTables renderer
    table2 = TableDisplay(headers, data, renderer=DataTablesRenderer())



# from edsl.utilities.interface import (
#     print_dict_as_html_table,
#     print_dict_with_rich,
#     print_list_of_dicts_as_html_table,
#     print_table_with_rich,
#     print_public_methods_with_doc,
#     print_list_of_dicts_as_markdown_table,
# )

# from edsl.utilities.utilities import (
#     create_valid_var_name,
#     dict_to_html,
#     hash_value,
#     HTMLSnippet,
#     is_notebook,
#     is_gzipped,
#     is_valid_variable_name,
#     random_string,
#     repair_json,
#     shorten_string,
#     time_all_functions,
# )



models_available = {
    "openai": [
        "gpt-3.5-turbo-1106",
        "gpt-4-0125-preview",
        "gpt-4-turbo-preview",
        "gpt-3.5-turbo-16k",
        "gpt-4-1106-preview",
        "gpt-4-turbo-2024-04-09",
        "gpt-3.5-turbo-16k-0613",
        "gpt-4o-2024-05-13",
        "gpt-4-turbo",
        "gpt-3.5-turbo-0613",
        "gpt-4",
        "gpt-4-0613",
        "gpt-3.5-turbo-0125",
        "gpt-3.5-turbo",
        "gpt-3.5-turbo-instruct",
        "gpt-3.5-turbo-instruct-0914",
        "gpt-3.5-turbo-0301",
        "gpt-4-vision-preview",
        "gpt-4-1106-vision-preview",
        "gpt-4o",
    ],
    "anthropic": [
        "claude-3-5-sonnet-20240620",
        "claude-3-opus-20240229",
        "claude-3-sonnet-20240229",
        "claude-3-haiku-20240307",
    ],
    "deep_infra": [
        "meta-llama/Llama-2-13b-chat-hf",
        "mistralai/Mixtral-8x22B-Instruct-v0.1",
        "Gryphe/MythoMax-L2-13b-turbo",
        "mistralai/Mistral-7B-Instruct-v0.1",
        "Austism/chronos-hermes-13b-v2",
        "meta-llama/Llama-2-70b-chat-hf",
        "mistralai/Mistral-7B-Instruct-v0.3",
        "meta-llama/Llama-2-7b-chat-hf",
        "Qwen/Qwen2-72B-Instruct",
        "HuggingFaceH4/zephyr-orpo-141b-A35b-v0.1",
        "cognitivecomputations/dolphin-2.6-mixtral-8x7b",
        "bigcode/starcoder2-15b",
        "microsoft/WizardLM-2-8x22B",
        "codellama/CodeLlama-70b-Instruct-hf",
        "Gryphe/MythoMax-L2-13b",
        "microsoft/WizardLM-2-7B",
        "01-ai/Yi-34B-Chat",
        "bigcode/starcoder2-15b-instruct-v0.1",
        "mistralai/Mixtral-8x7B-Instruct-v0.1",
        "openchat/openchat-3.6-8b",
        "meta-llama/Meta-Llama-3-8B-Instruct",
        "microsoft/Phi-3-medium-4k-instruct",
        "Phind/Phind-CodeLlama-34B-v2",
        "google/codegemma-7b-it",
        "mistralai/Mistral-7B-Instruct-v0.2",
        "deepinfra/airoboros-70b",
        "mistralai/Mixtral-8x22B-v0.1",
        "llava-hf/llava-1.5-7b-hf",
        "codellama/CodeLlama-34b-Instruct-hf",
        "google/gemma-1.1-7b-it",
        "lizpreciatior/lzlv_70b_fp16_hf",
        "databricks/dbrx-instruct",
        "nvidia/Nemotron-4-340B-Instruct",
        "Qwen/Qwen2-7B-Instruct",
        "meta-llama/Meta-Llama-3-70B-Instruct",
        "openchat/openchat_3.5",
    ],
    "google": [
        "gemini-1.0-pro",
        "gemini-1.0-pro-001",
        "gemini-1.0-pro-latest",
        "gemini-1.0-pro-vision-latest",
        "gemini-1.5-flash",
        "gemini-1.5-flash-001",
        "gemini-1.5-flash-001-tuning",
        "gemini-1.5-flash-002",
        "gemini-1.5-flash-8b",
        "gemini-1.5-flash-8b-001",
        "gemini-1.5-flash-8b-exp-0827",
        "gemini-1.5-flash-8b-exp-0924",
        "gemini-1.5-flash-8b-latest",
        "gemini-1.5-flash-exp-0827",
        "gemini-1.5-flash-latest",
        "gemini-1.5-pro",
        "gemini-1.5-pro-001",
        "gemini-1.5-pro-002",
        "gemini-1.5-pro-exp-0801",
        "gemini-1.5-pro-exp-0827",
        "gemini-1.5-pro-latest",
        "gemini-pro",
        "gemini-pro-vision",
    ],
    "bedrock": [
        "amazon.titan-tg1-large",
        "amazon.titan-text-lite-v1",
        "amazon.titan-text-express-v1",
        "anthropic.claude-instant-v1",
        "anthropic.claude-v2:1",
        "anthropic.claude-v2",
        "anthropic.claude-3-sonnet-20240229-v1:0",
        "anthropic.claude-3-haiku-20240307-v1:0",
        "anthropic.claude-3-opus-20240229-v1:0",
        "anthropic.claude-3-5-sonnet-20240620-v1:0",
        "cohere.command-text-v14",
        "cohere.command-r-v1:0",
        "cohere.command-r-plus-v1:0",
        "cohere.command-light-text-v14",
        "meta.llama3-8b-instruct-v1:0",
        "meta.llama3-70b-instruct-v1:0",
        "meta.llama3-1-8b-instruct-v1:0",
        "meta.llama3-1-70b-instruct-v1:0",
        "meta.llama3-1-405b-instruct-v1:0",
        "mistral.mistral-7b-instruct-v0:2",
        "mistral.mixtral-8x7b-instruct-v0:1",
        "mistral.mistral-large-2402-v1:0",
        "mistral.mistral-large-2407-v1:0",
    ],
}



from typing import Optional, List, TYPE_CHECKING
from collections import UserList

from edsl.Base import Base
from edsl.language_models.model import Model

#
from edsl.utilities.remove_edsl_version import remove_edsl_version
from edsl.utilities.is_valid_variable_name import is_valid_variable_name

if TYPE_CHECKING:
    from edsl.inference_services.data_structures import AvailableModels
    from edsl.language_models import LanguageModel


class ModelList(Base, UserList):
    __documentation__ = """https://docs.expectedparrot.com/en/latest/language_models.html#module-edsl.language_models.ModelList"""

    def __init__(self, data: Optional["LanguageModel"] = None):
        """Initialize the ScenarioList class.

        >>> from edsl import Model
        >>> m = ModelList(Model.available())

        """
        if data is not None:
            super().__init__(data)
        else:
            super().__init__([])

    @property
    def names(self):
        """

        >>> ModelList.example().names
        {'...'}
        """
        return set([model.model for model in self])

    def __repr__(self):
        return f"ModelList({super().__repr__()})"

    def _summary(self):
        return {"models": len(self)}

    def __hash__(self):
        """Return a hash of the ModelList. This is used for comparison of ModelLists.

        >>> isinstance(hash(Model()), int)
        True

        """
        from edsl.utilities.utilities import dict_hash

        return dict_hash(self.to_dict(sort=True, add_edsl_version=False))

    def to_scenario_list(self):
        from edsl.scenarios.ScenarioList import ScenarioList
        from edsl.scenarios.Scenario import Scenario

        sl = ScenarioList()
        for model in self:
            d = {"model": model.model, "inference_service": model._inference_service_}
            d.update(model.parameters)
            sl.append(Scenario(d))
        return sl

    def tree(self, node_list: Optional[List[str]] = None):
        return self.to_scenario_list().tree(node_list)

    def table(
        self,
        *fields,
        tablefmt: Optional[str] = None,
        pretty_labels: Optional[dict] = None,
    ):
        """
        >>> ModelList.example().table('model')
        model
        -------
        gpt-4o
        gpt-4o
        gpt-4o
        """
        return (
            self.to_scenario_list()
            .to_dataset()
            .table(*fields, tablefmt=tablefmt, pretty_labels=pretty_labels)
        )

    def to_list(self) -> list:
        return self.to_scenario_list().to_list()

    def to_dict(self, sort=False, add_edsl_version=True):
        if sort:
            model_list = sorted([model for model in self], key=lambda x: hash(x))
            d = {
                "models": [
                    model.to_dict(add_edsl_version=add_edsl_version)
                    for model in model_list
                ]
            }
        else:
            d = {
                "models": [
                    model.to_dict(add_edsl_version=add_edsl_version) for model in self
                ]
            }
        if add_edsl_version:
            from edsl import __version__

            d["edsl_version"] = __version__
            d["edsl_class_name"] = "ModelList"

        return d

    @classmethod
    def from_names(self, *args, **kwargs):
        """A a model list from a list of names"""
        if len(args) == 1 and isinstance(args[0], list):
            args = args[0]
        return ModelList([Model(model_name, **kwargs) for model_name in args])

    @classmethod
    def from_available_models(self, available_models_list: "AvailableModels"):
        """Create a ModelList from an AvailableModels object"""
        return ModelList(
            [
                Model(model.model_name, service_name=model.service_name)
                for model in available_models_list
            ]
        )

    @classmethod
    @remove_edsl_version
    def from_dict(cls, data):
        """
        Create a ModelList from a dictionary.

        >>> newm = ModelList.from_dict(ModelList.example().to_dict())
        >>> assert ModelList.example() == newm
        """
        from edsl.language_models.LanguageModel import LanguageModel

        return cls(data=[LanguageModel.from_dict(model) for model in data["models"]])

    def code(self):
        pass

    @classmethod
    def example(cls, randomize: bool = False) -> "ModelList":
        """
        Returns an example ModelList instance.

        :param randomize: If True, uses Model's randomize method.
        """

        return cls([Model.example(randomize) for _ in range(3)])


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)
